/* This file is Copyright 2001 Level Control Systems.  See the included LICENSE.txt file for details. */

#ifndef FILE_DESCRIPTOR_DATA_IO
#define FILE_DESCRIPTOR_DATA_IO

#include "support/PortableMacros.h"

#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include "dataio/PortableDataIO.h"

/**
 *  Data I/O to and from a file descriptor (useful for talking to Linux device drivers and the like)
 */
class FileDescriptorDataIO : public PortableDataIO
{
public:
   /**
    *  Constructor.
    *  @param fd The file descriptor to use.  Becomes property of this FileDescriptorDataIO object.
    *  @blockingIO determines whether to use blocking or non-blocking I/O.
    *  If you will be using this object with a AbstractMessageIOGateway,
    *  and/or select(), then it's usually better to set blocking to false.
    */
   FileDescriptorDataIO(int fd, bool blocking) : _fd(fd)
   {
      SetBlockingIOEnabled(blocking);
   }

   /** Destructor.
    *  close()'s the held file descriptor.
    */
   virtual ~FileDescriptorDataIO() 
   {
      close(_fd);
   }

   /** Reads bytes from the file descriptor and places them into (buffer).
    *  @param buffer Buffer to write the bytes into
    *  @param size Number of bytes in the buffer.
    *  @return Number of bytes read, or -1 on error.
    *  @see PortableDataIO::Read()
    */
   virtual ssize_t Read(void * buffer, size_t size)  
   {
      ssize_t ret = read(_fd, buffer, size);
      return CalculateReturnValue(ret);
   }

   /**
    * Reads bytes from (buffer) and sends them out to the file descriptor.
    *  @param buffer Buffer to read the bytes from.
    *  @param size Number of bytes in the buffer.
    *  @return Number of bytes writte, or -1 on error.
    *  @see PortableDataIO::Write()    
    */
   virtual ssize_t Write(const void * buffer, size_t size)
   {
      ssize_t ret = write(_fd, buffer, size);
      return CalculateReturnValue(ret);
   }

   /**
    *  Implemented as a no-op (I don't believe file descriptors need flushing?)
    */
   virtual void FlushOutput()
   {
      // empty
   }
   
   /**
    * Enables or diables blocking I/O on this file descriptor.
    * If this object is to be used by an AbstractMessageIOGateway,
    * then non-blocking I/O is usually better to use.
    * @param blocking If true, file descriptor is set to blocking I/O mode.  Otherwise, non-blocking I/O.
    * @return B_NO_ERROR on success, B_ERROR on error.
    */
   status_t SetBlockingIOEnabled(bool blocking)
   {
      if (fcntl(_fd, F_SETFL, blocking ? 0 : O_NONBLOCK) == 0)
      {
         _blocking = blocking;
         return B_NO_ERROR;
      }
      else
      {
         perror("FileDescriptorDataIO:SetBlockingIO failed");
         return B_ERROR;
      }
   }

   /**
    * Releases control of the contained file descriptor to the calling code.
    * After this method returns, this object no longer owns or can
    * use or close the file descriptor descriptor it once held.
    */
   void ReleaseFileDescriptor() {_fd = -1;}

   /**
    * Returns the file descriptor held by this object, or
    * -1 if there is none.
    */
   int GetFileDescriptor() const {return _fd;}

private:
   int CalculateReturnValue(int ret)
   {
      int retForBlocking = ((ret == 0) ? -1 : ret);
      return (_blocking) ? retForBlocking : ret;
   }

   int _fd;
   bool _blocking;
};

#endif
