/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */  

#include <new.h>
#include <stdio.h>
#include "queue/PortableQueue.h"
#include "message/PortableMessage.h"
#include "support/PortableByteOrder.h"
#include "support/PortableMacros.h"

static void ClearMsgFunc(PortableMessage * msg, void * userData);
void ClearMsgFunc(PortableMessage * msg, void * /*userData*/)
{
   msg->MakeEmpty();
}

PortableMessageRef::ItemPool _messagePool(100, ClearMsgFunc);
PortableMessageRef::RefPool _messageRefPool;

PortableMessageRef::ItemPool * GetMessagePool() {return &_messagePool;}
PortableMessageRef::RefPool * GetMessageRefPool() {return &_messageRefPool;}

PortableMessageRef::ItemPool * _defaultMessagePool = NULL;
PortableMessageRef::RefPool * _defaultMessageRefPool = NULL;

PortableMessage * GetMessageFromPool(uint32 what)
{
   PortableMessage * ret = _messagePool.GetObject();
   if (ret) ret->what = what;
   return ret;
}

PortableMessage * GetMessageFromPool(const PortableMessage & copyMe)
{
   PortableMessage * ret = _messagePool.GetObject();
   if (ret) *ret = copyMe;
   return ret;
}

PortableMessageFieldNameIterator :: PortableMessageFieldNameIterator() 
   : _owner(NULL), _iter(NULL), _typeCode(B_ANY_TYPE)
{
   // empty
}

PortableMessageFieldNameIterator :: ~PortableMessageFieldNameIterator() 
{
   delete _iter;
}

PortableMessageFieldNameIterator :: PortableMessageFieldNameIterator(const PortableMessageFieldNameIterator & copyMe) : _owner(copyMe._owner), _iter(NULL)
{
   *this = copyMe;
}

PortableMessageFieldNameIterator & 
PortableMessageFieldNameIterator :: operator = (const PortableMessageFieldNameIterator & rhs)
{
   if (this != &rhs)
   {
      delete _iter;
      _iter = NULL;
      _owner = rhs._owner;
      _typeCode = rhs._typeCode;
      if (rhs._iter) 
      {
         _iter = newnothrow _eitr(*rhs._iter);
         if (_iter == NULL) WARN_OUT_OF_MEMORY;
      }
   }
   return *this;
}

bool PortableMessageFieldNameIterator :: HasMoreFieldNames() const 
{
   return (PeekNextFieldName() != NULL);
}

bool PortableMessageFieldNameIterator :: GetNextFieldName(PortableString & name)
{
   const char * ret = GetNextFieldName();
   if (ret)
   {
      name = ret;
      return true;
   }
   return false;
}

bool PortableMessageFieldNameIterator :: PeekNextFieldName(PortableString & name) const
{
   const char * ret = PeekNextFieldName();
   if (ret)
   {
      name = ret;
      return true;
   }
   return false;
}

const char * PortableMessageFieldNameIterator :: GetNextFieldName()
{
   return _iter ? _owner->GetNextFieldName(_iter, _typeCode) : NULL;
}

const char * PortableMessageFieldNameIterator :: PeekNextFieldName() const
{
   return _iter ? _owner->PeekNextFieldName(_iter, _typeCode) : NULL;
}


/* This class is for the private use of the PortableMessage classe only! 
 * It represents one "name" of the message, which can in turn represent 1 or more
 * data items of a single type. 
 */
class AbstractDataArray : public PortableFlattenable 
{
public:
   // Should add the given item to our internal array.
   virtual status_t AddDataItem(const void * data, size_t size) = 0;

   // Should remove the (index)'th item from our internal array.
   virtual status_t RemoveDataItem(int32 index) = 0;

   // Prepends the given data item to the beginning of our array.
   virtual status_t PrependDataItem(const void * data, size_t size) = 0;

   // Clears the array
   virtual void MakeEmpty() = 0;

   // Sets (setDataLoc) to point to the (index)'th item in our array.
   // is not guaranteed to remain valid after this object is modified.
   virtual status_t FindDataItem(int32 index, const void ** setDataLoc) const = 0;

   // Should replace the (index)'th data item in the array with (data).
   virtual status_t ReplaceDataItem(int32 index, const void * data, size_t size) = 0;

   // Returns the size (in bytes) of the item in the (index)'th slot.
   virtual size_t GetItemSize(int32 index) const = 0;

   // Returns the number of items currently in the array
   virtual uint32 CountItems() const = 0;

   // Returns true iff all elements in the array have the same size
   virtual bool ElementsAreFixedSize() const = 0;

   // PortableFlattenable interface
   virtual bool IsFixedSize() const {return false;}

   // returns a separate (deep) copy of this array
   virtual AbstractDataArray * Clone() const = 0;

   // For debugging:  should dump our contents to stdout
   virtual void PrintToStream() = 0;
};

template <class DataType> class FixedSizeDataArray : public AbstractDataArray
{
public:
   FixedSizeDataArray() {/* empty */}
   virtual ~FixedSizeDataArray() {/* empty */}

   virtual uint32 CountItems() const {return _data.CountItems();}

   virtual void MakeEmpty() {_data.MakeEmpty();}

   virtual status_t AddDataItem(const void * item, size_t size)
   {
      if (size == sizeof(DataType))
      {
         if (item) return (_data.AddTail(*((DataType *)item))) ? B_NO_ERROR : B_ERROR;
         else
         {
            DataType defaultItem;
            return (_data.AddTail(defaultItem)) ? B_NO_ERROR : B_ERROR;
         }
      }
      return B_ERROR;
   }

   virtual status_t PrependDataItem(const void * item, size_t size)
   {
      if (size == sizeof(DataType))
      {
         if (item) return (_data.AddHead(*((DataType *)item))) ? B_NO_ERROR : B_ERROR;
         else
         {
            DataType defaultItem;
            return (_data.AddHead(defaultItem)) ? B_NO_ERROR : B_ERROR;
         }
      }
      return B_ERROR;
   }

   virtual size_t GetItemSize(int32 /*index*/) const {return sizeof(DataType);}

   virtual status_t RemoveDataItem(int32 index) 
   {
      return _data.RemoveItemAt(index) ? B_NO_ERROR : B_ERROR;
   }

   virtual status_t FindDataItem(int32 index, const void ** setDataLoc) const
   {
      if ((index >= 0)&&(index < (int32) _data.CountItems()))
      {
         *setDataLoc = _data.GetItemPointer(index);
         return (*setDataLoc) ? B_NO_ERROR : B_ERROR;
      }
      return B_ERROR;
   }

   virtual status_t ReplaceDataItem(int32 index, const void * data, size_t size)
   {
      return ((size == sizeof(DataType))&&(_data.ReplaceItemAt(index, *((DataType *)data)))) ? B_NO_ERROR : B_ERROR;
   }

   virtual bool ElementsAreFixedSize() const {return true;}

   DataType ItemAt(int i) const  
   {
      DataType ret;
      _data.GetItemAt(i, ret);
      return ret;
   }

protected:
   FixedSizeDataArray<DataType> & operator=(const FixedSizeDataArray<DataType> &rhs) 
   {
      if (this != &rhs) _data = rhs._data;
      return *this;
   }

   PortableQueue<DataType> _data;
};

// an array of flattened objects, where all flattened objects are of the same size
template <class DataType> class FixedSizeFlatObjectArray : public FixedSizeDataArray<DataType>
{
public:
   FixedSizeFlatObjectArray() 
   {
      DataType temp;
      _typeCode = temp.TypeCode();
      _flatItemSize = temp.FlattenedSize();
   }

   virtual ~FixedSizeFlatObjectArray() 
   {
      // empty
   }

   virtual void Flatten(uint8 *buffer) const
   {
      int numItems = _data.CountItems(); 
      for (int i=0; i<numItems; i++) 
      {
         _data.GetItemPointer(i)->Flatten(buffer);
         buffer += _flatItemSize;
      }
   }

   // PortableFlattenable interface
   virtual size_t FlattenedSize() const {return _data.CountItems() * _flatItemSize;}

   virtual status_t Unflatten(const uint8 *buffer, size_t numBytes)
   {
      _data.MakeEmpty();
      if (numBytes % _flatItemSize) return B_ERROR;  // length must be an even multiple of item size, or something's wrong!

      DataType temp;
      int numItems = numBytes / _flatItemSize;
      if (_data.EnsureSize(numItems))
      {
         for (int i=0; i<numItems; i++)
         {
            if (temp.Unflatten(buffer, _flatItemSize) != B_NO_ERROR) return B_ERROR;
            _data.AddTail(temp);
            buffer += _flatItemSize;
         }
         return B_NO_ERROR;
      }
      else return B_ERROR;
   }

   virtual type_code TypeCode() const {return _typeCode;}

   virtual AbstractDataArray * Clone() const 
   {
      FixedSizeFlatObjectArray<DataType> * c = newnothrow FixedSizeFlatObjectArray<DataType>;
      if (c == NULL) {WARN_OUT_OF_MEMORY; return NULL;}
      *c = *this;
      return c;
   }

protected:
   virtual void PrintToStream()
   {
      int numItems = CountItems();
      for (int i=0; i<numItems; i++)
      {
         printf("    %i. ", i);
         _data.GetItemPointer(i)->PrintToStream(); 
      }
   }

private:
   type_code _typeCode;
   size_t _flatItemSize;
};

typedef FixedSizeFlatObjectArray<PortablePoint> PointDataArray;
typedef FixedSizeFlatObjectArray<PortableRect> RectDataArray;

/* This class handles storage of all the primitive numeric types for us. */
template <class DataType> class PrimitiveTypeDataArray : public FixedSizeDataArray<DataType> 
{
public:
   PrimitiveTypeDataArray() {/* empty */}
   virtual ~PrimitiveTypeDataArray() {/* empty */}

   virtual void Flatten(uint8 *buffer) const
   {
      DataType * dBuf = (DataType *) buffer;
      int numItems = _data.CountItems(); 
      for (int i=0; i<numItems; i++) ConvertToNetworkByteOrder(&dBuf[i], _data.GetItemPointer(i));
   }

   virtual status_t Unflatten(const uint8 *buffer, size_t numBytes)
   {
      _data.MakeEmpty();
      if (numBytes % sizeof(DataType)) return B_ERROR;  // length must be an even multiple of item size, or something's wrong!

      int numItems = numBytes / sizeof(DataType);
      if (_data.EnsureSize(numItems))
      {
         DataType * dBuf = (DataType *) buffer;
         for (int i=0; i<numItems; i++)
         {
            DataType next;
            ConvertFromNetworkByteOrder(&next, &dBuf[i]);
            _data.AddTail(next);
         }
         return B_NO_ERROR;
      }
      else return B_ERROR;
   }

   // PortableFlattenable interface
   virtual size_t FlattenedSize() const {return _data.CountItems() * sizeof(DataType);}

protected:
   virtual void ConvertToNetworkByteOrder(void * writeToHere, const void * readFromHere) const = 0;
   virtual void ConvertFromNetworkByteOrder(void * writeToHere, const void * readFromHere) const = 0;

   // For debugging, mostly:  writes a string representation of the (index)'th entry into (buf)
   virtual const char * GetFormatString() const = 0;

   virtual void PrintToStream()
   {
      char temp[100];
      int numItems = CountItems();
      for (int i=0; i<numItems; i++)
      {
         sprintf(temp, GetFormatString(), ItemAt(i));
         printf("    %i. [%s]\n", i, temp);
      }
   }
};

class Int8DataArray : public PrimitiveTypeDataArray<int8> 
{
public:
   Int8DataArray() {/* empty */}
   virtual ~Int8DataArray() {/* empty */}

   virtual type_code TypeCode() const {return B_INT8_TYPE;}

   virtual const char * GetFormatString() const {return "%i";}
 
   virtual AbstractDataArray * Clone() const 
   {
      Int8DataArray * c = newnothrow Int8DataArray;
      if (c == NULL) {WARN_OUT_OF_MEMORY; return NULL;}
      *c = *this;
      return c;
   }

protected:
   virtual void ConvertToNetworkByteOrder(void * writeToHere, const void * readFromHere) const
   {
      *((uint8*)writeToHere) = *((uint8*)readFromHere);  // no translation required, really
   }

   virtual void ConvertFromNetworkByteOrder(void * writeToHere, const void * readFromHere) const
   {
      *((uint8*)writeToHere) = *((uint8*)readFromHere);  // no translation required, really
   }
};

class BoolDataArray : public Int8DataArray
{
public:
   BoolDataArray() {/* empty */}
   virtual ~BoolDataArray() {/* empty */}

   virtual type_code TypeCode() const {return B_BOOL_TYPE;}

   virtual const char * GetFormatString() const {return "%i";}

   virtual AbstractDataArray * Clone() const 
   {
      BoolDataArray * c = newnothrow BoolDataArray;
      if (c == NULL) {WARN_OUT_OF_MEMORY; return NULL;}
      *c = *this;
      return c;
   }

protected:
   virtual void ConvertToNetworkByteOrder(void * writeToHere, const void * readFromHere) const
   {
      *((uint8*)writeToHere) = *((uint8*)readFromHere);  // no translation required, really
   }

   virtual void ConvertFromNetworkByteOrder(void * writeToHere, const void * readFromHere) const
   {
      *((uint8*)writeToHere) = *((uint8*)readFromHere);  // no translation required, really
   }
};


class Int16DataArray : public PrimitiveTypeDataArray<int16> 
{
public:
   Int16DataArray() {/* empty */}
   virtual ~Int16DataArray() {/* empty */}

   virtual type_code TypeCode() const {return B_INT16_TYPE;}

   virtual const char * GetFormatString() const {return "%i";}

   virtual AbstractDataArray * Clone() const 
   {
      Int16DataArray * c = newnothrow Int16DataArray;
      if (c == NULL) {WARN_OUT_OF_MEMORY; return NULL;}
      *c = *this;
      return c;
   }

protected:
   virtual void ConvertToNetworkByteOrder(void * writeToHere, const void * readFromHere) const
   {
      *((uint16*)writeToHere) = B_HOST_TO_LENDIAN_INT16(*((uint16*)readFromHere));  
   }

   virtual void ConvertFromNetworkByteOrder(void * writeToHere, const void * readFromHere) const
   {
      *((uint16*)writeToHere) = B_LENDIAN_TO_HOST_INT16(*((uint16*)readFromHere)); 
   }
};

class Int32DataArray : public PrimitiveTypeDataArray<int32> 
{
public:
   Int32DataArray() {/* empty */}
   virtual ~Int32DataArray() {/* empty */}

   virtual type_code TypeCode() const {return B_INT32_TYPE;}

   virtual const char * GetFormatString() const {return "%li";}

   virtual AbstractDataArray * Clone() const 
   {
      Int32DataArray * c = newnothrow Int32DataArray;
      if (c == NULL) {WARN_OUT_OF_MEMORY; return NULL;}
      *c = *this;
      return c;
   }

protected:
   virtual void ConvertToNetworkByteOrder(void * writeToHere, const void * readFromHere) const
   {
      *((uint32*)writeToHere) = B_HOST_TO_LENDIAN_INT32(*((uint32*)readFromHere));  
   }

   virtual void ConvertFromNetworkByteOrder(void * writeToHere, const void * readFromHere) const
   {
      *((uint32*)writeToHere) = B_LENDIAN_TO_HOST_INT32(*((uint32*)readFromHere)); 
   }
};

class Int64DataArray : public PrimitiveTypeDataArray<int64> 
{
public:
   Int64DataArray() {/* empty */}
   virtual ~Int64DataArray() {/* empty */}

   virtual type_code TypeCode() const {return B_INT64_TYPE;}

   virtual const char * GetFormatString() const {return "%Li";}

   virtual AbstractDataArray * Clone() const 
   {
      Int64DataArray * c = newnothrow Int64DataArray;
      if (c == NULL) {WARN_OUT_OF_MEMORY; return NULL;}
      *c = *this;
      return c;
   }

protected:
   virtual void ConvertToNetworkByteOrder(void * writeToHere, const void * readFromHere) const
   {
      *((uint64*)writeToHere) = B_HOST_TO_LENDIAN_INT64(*((uint64*)readFromHere));  
   }

   virtual void ConvertFromNetworkByteOrder(void * writeToHere, const void * readFromHere) const
   {
      *((uint64*)writeToHere) = B_LENDIAN_TO_HOST_INT64(*((uint64*)readFromHere)); 
   }
};

class FloatDataArray : public PrimitiveTypeDataArray<float> 
{
public:
   FloatDataArray() {/* empty */}
   virtual ~FloatDataArray() {/* empty */}

   virtual type_code TypeCode() const {return B_FLOAT_TYPE;}

   virtual const char * GetFormatString() const {return "%f";}

   virtual AbstractDataArray * Clone() const 
   {
      FloatDataArray * c = newnothrow FloatDataArray;
      if (c == NULL) {WARN_OUT_OF_MEMORY; return NULL;}
      *c = *this;
      return c;
   }

protected:
   virtual void ConvertToNetworkByteOrder(void * writeToHere, const void * readFromHere) const
   {
      *((float*)writeToHere) = B_HOST_TO_LENDIAN_FLOAT(*((float*)readFromHere));  
   }

   virtual void ConvertFromNetworkByteOrder(void * writeToHere, const void * readFromHere) const
   {
      *((float*)writeToHere) = B_LENDIAN_TO_HOST_FLOAT(*((float*)readFromHere)); 
   }
};

class DoubleDataArray : public PrimitiveTypeDataArray<double> 
{
public:
   DoubleDataArray() {/* empty */}
   virtual ~DoubleDataArray() {/* empty */}

   virtual type_code TypeCode() const {return B_DOUBLE_TYPE;}

   virtual const char * GetFormatString() const {return "%lf";}

   virtual AbstractDataArray * Clone() const 
   {
      DoubleDataArray * c = newnothrow DoubleDataArray;
      if (c == NULL) {WARN_OUT_OF_MEMORY; return NULL;}
      *c = *this;
      return c;
   }

protected:
   virtual void ConvertToNetworkByteOrder(void * writeToHere, const void * readFromHere) const
   {
      *((double*)writeToHere) = B_HOST_TO_LENDIAN_DOUBLE(*((double*)readFromHere));  
   }

   virtual void ConvertFromNetworkByteOrder(void * writeToHere, const void * readFromHere) const
   {
      *((double*)writeToHere) = B_LENDIAN_TO_HOST_DOUBLE(*((double*)readFromHere)); 
   }
};

class PointerDataArray : public PrimitiveTypeDataArray<void *> 
{
public:
   PointerDataArray() {/* empty */}
   virtual ~PointerDataArray() {/* empty */}

   virtual type_code TypeCode() const {return B_POINTER_TYPE;}

   virtual const char * GetFormatString() const {return "%p";}

   virtual AbstractDataArray * Clone() const 
   {
      PointerDataArray * c = newnothrow PointerDataArray;
      if (c == NULL) {WARN_OUT_OF_MEMORY; return NULL;}
      *c = *this;
      return c;
   }

protected:
   virtual void ConvertToNetworkByteOrder(void * writeToHere, const void * readFromHere) const
   {
      // This will break on non 32-bit systems, but then again it's a bad idea to
      // be serializing pointers anyway..
      *((uint32*)writeToHere) = B_HOST_TO_LENDIAN_INT32(*((uint32*)readFromHere));  
   }

   virtual void ConvertFromNetworkByteOrder(void * writeToHere, const void * readFromHere) const
   {
      // This will break on non 32-bit systems, but then again it's a bad idea to
      // be serializing pointers anyway..
      *((uint32*)writeToHere) = B_LENDIAN_TO_HOST_INT32(*((uint32*)readFromHere)); 
   }
};


 /* objects, and deletes them as necessary to avoid memory leaks.
 * NOTE:  AddDataItem() here expects (data) to have been allocated with the new operator, and will delete [] data
 *        when removing it from the collection!!!
 */
class ByteSequenceDataArray : public PrimitiveTypeDataArray<uint8 *>
{
public:
   ByteSequenceDataArray() {/* empty */}
   virtual ~ByteSequenceDataArray() {MakeEmpty();} 

   ByteSequenceDataArray(type_code typeCode) 
      : _totalHeldBytes(0), _typeCode(typeCode) {/* empty */}

   ByteSequenceDataArray(const ByteSequenceDataArray & copyMe)
      : PrimitiveTypeDataArray<uint8*>(), _totalHeldBytes(0) {*this = copyMe;}

   virtual void MakeEmpty() 
   {
      // Since we own all the items stored in the array, we must delete them to avoid memory leaks
      for (int i=CountItems()-1; i>=0; i--) RemoveDataItem(i);
   }      

   virtual status_t FindDataItem(int32 index, const void ** setDataLoc) const
   {
      const uint8 ** getAddress;
      status_t ret = PrimitiveTypeDataArray<uint8 *>::FindDataItem(index, (const void **)&getAddress);
      if (ret != B_NO_ERROR) return ret;
      *setDataLoc = *getAddress;  // they want pointer to data, NOT pointer to the pointer!
      return B_NO_ERROR;
   }

   virtual bool ElementsAreFixedSize() const {return false;}

   // Should add the given item to our internal array.
   virtual status_t AddDataItem(const void * data, size_t size) 
   {
      bool alloced = false;
      if (size <= 0) return B_NO_ERROR;  // we don't support 0-length data arrays, sorry

      if (data == NULL)
      {
         data = newnothrow uint8[size];
         if (data) alloced = true;
         else
         {
            WARN_OUT_OF_MEMORY;
            return B_ERROR;
         }
      }

      status_t ret = PrimitiveTypeDataArray<uint8*>::AddDataItem(&data, sizeof(uint8 *));
      if (ret == B_NO_ERROR) 
      {
         ret = _sizeArray.AddDataItem(&size, sizeof(size));
         if (ret == B_NO_ERROR) _totalHeldBytes += size;
         else 
         {
            PrimitiveTypeDataArray<uint8 *>::RemoveDataItem(CountItems()-1);  // rollback
            if (alloced) delete [] ((uint8*)data);
         }
      }
      return ret;
   }

   virtual status_t PrependDataItem(const void * data, size_t size)
   {
      bool alloced = false;
      if (size <= 0) return B_NO_ERROR;  // we don't support 0-length data arrays, sorry

      if (data == NULL)
      {
         data = newnothrow uint8[size];
         if (data) alloced = true;
         else
         {
            WARN_OUT_OF_MEMORY;
            return B_ERROR;
         }
      }

      status_t ret = PrimitiveTypeDataArray<uint8*>::PrependDataItem(&data, sizeof(uint8 *));
      if (ret == B_NO_ERROR) 
      {
         ret = _sizeArray.PrependDataItem(&size, sizeof(size));
         if (ret == B_NO_ERROR) _totalHeldBytes += size;
         else 
         {
            PrimitiveTypeDataArray<uint8 *>::RemoveDataItem(0);  // rollback
            if (alloced) delete [] ((uint8*)data);
         }
      }
      return ret;
   }

   virtual status_t RemoveDataItem(int32 index)
   {
      const uint8 * oldData;
      status_t ret = FindDataItem(index, (const void **) &oldData);
      if (ret == B_NO_ERROR)
      {
         size_t oldSize = GetItemSize(index);
         ret = PrimitiveTypeDataArray<uint8 *>::RemoveDataItem(index);
         if (ret == B_NO_ERROR) 
         {
            delete [] oldData;
            _sizeArray.RemoveDataItem(index);
            _totalHeldBytes -= oldSize;
         }
      }
      return ret;
   }

   virtual size_t GetItemSize(int32 index) const
   {
      const size_t * size = 0;
      return (_sizeArray.FindDataItem(index, (const void **) &size) == B_NO_ERROR) ? *size : 0L;
   }

   virtual status_t ReplaceDataItem(int32 index, const void * data, size_t numBytes)
   {
      const uint8 * oldData;
      status_t ret = FindDataItem(index, (const void **) &oldData);
      if (ret == B_NO_ERROR) 
      {
         size_t oldSize = GetItemSize(index);
         ret = PrimitiveTypeDataArray<uint8 *>::ReplaceDataItem(index, &data, sizeof(uint8 *));
         if (ret == B_NO_ERROR) 
         {
            (void) _sizeArray.ReplaceDataItem(index, &numBytes, sizeof(size_t));
            delete [] oldData;
            _totalHeldBytes += (numBytes - oldSize);
         }
      }
      return ret;
   }

   // PortableFlattenable interface
   virtual type_code TypeCode() const {return _typeCode;}
   virtual size_t FlattenedSize() const {return sizeof(uint32) + _totalHeldBytes + CountItems()*sizeof(uint32);}

   virtual void Flatten(uint8 *buffer) const
   {
      // Format:  0. number of entries (4 bytes)
      //          1. entry size in bytes (4 bytes)
      //          2. entry data (n bytes)
      //          (repeat 1. and 2. as necessary)
      uint32 numElements = CountItems();
      uint32 networkByteOrder = B_HOST_TO_LENDIAN_INT32(numElements);
      size_t writeOffset = 0;

      WriteData(buffer, &writeOffset, &networkByteOrder, sizeof(networkByteOrder));

      for (uint32 i=0; i<numElements; i++)
      {
         // write element size
         size_t nextElementBytes = GetItemSize(i);
         networkByteOrder = B_HOST_TO_LENDIAN_INT32(nextElementBytes);
         WriteData(buffer, &writeOffset, &networkByteOrder, sizeof(networkByteOrder));

         // write element data
         const uint8 * nextData;
         (void) FindDataItem(i, (const void **) &nextData);
         WriteData(buffer, &writeOffset, nextData, nextElementBytes);
      }
   }

   virtual status_t Unflatten(const uint8 *buffer, size_t inputBufferBytes)
   {
      MakeEmpty();

      uint32 networkByteOrder;
      size_t readOffset = 0;

      if (ReadData(buffer, inputBufferBytes, &readOffset, &networkByteOrder, sizeof(networkByteOrder)) == false) return B_ERROR;
      size_t numElements = B_LENDIAN_TO_HOST_INT32(networkByteOrder);

      for (size_t i=0; i<numElements; i++) 
      {
         if (ReadData(buffer, inputBufferBytes, &readOffset, &networkByteOrder, sizeof(networkByteOrder)) == false) return B_ERROR;
         size_t elementSize = B_LENDIAN_TO_HOST_INT32(networkByteOrder);
         if (elementSize == 0) return B_ERROR;

         // read element data
         uint8 * data = newnothrow uint8[elementSize];
         if (data == NULL) {WARN_OUT_OF_MEMORY; return B_ERROR;}
         if ((ReadData(buffer, inputBufferBytes, &readOffset, data, elementSize) == false)||(AddDataItem(data, elementSize) != B_NO_ERROR)) 
         {
            delete [] data;
            return B_ERROR;
         }
      }
      return B_NO_ERROR;
   }

   virtual const char * GetFormatString() const {return "%p";}

   virtual AbstractDataArray * Clone() const 
   {
       ByteSequenceDataArray * c = newnothrow ByteSequenceDataArray(TypeCode());
       if (c == NULL) {WARN_OUT_OF_MEMORY; return NULL;}
       *c = *this;
       return c;
   }

   ByteSequenceDataArray & operator=(const ByteSequenceDataArray &rhs) 
   {
       if (this != &rhs)
       {
          MakeEmpty();
          _typeCode = rhs._typeCode;
          int numElements = rhs.CountItems();
          for (int i=0; i<numElements; i++)
          {
             const uint8 * nextData;
             if (rhs.FindDataItem(i, (const void **) &nextData) == B_NO_ERROR)
             {
                size_t itemSize = rhs.GetItemSize(i);
                uint8 * copyData = newnothrow uint8[itemSize];
                if (copyData)
                {
                   memcpy(copyData, nextData, itemSize);
                   if (AddDataItem(copyData, itemSize) != B_NO_ERROR) delete [] copyData;
                }
                else 
                {
                   WARN_OUT_OF_MEMORY;
                   break;
                }
             }
          }
       }
       return *this;
   }

   virtual void PrintToStream()
   {
      char bbuf[5];
      MakePrettyTypeCodeString(TypeCode(), bbuf);
      int numItems = CountItems();
      for (int i=0; i<numItems; i++)
      {
         uint32 s = _sizeArray.ItemAt(i);
         printf("    %i. [%lu bytes of data: ", i, s);
  
         const uint8 * data;
         if (FindDataItem(i, (const void **) &data) == B_NO_ERROR)
         {
            uint32 numBytesToPrint = (s < 10) ? s : 10;
            for (uint32 j=0; j<numBytesToPrint; j++) printf("%x%s", data[j], (j==numBytesToPrint-1)?"":" ");
         }
         printf("%s]\n", (s<10)?"":"...");
      }
   }

protected:
   virtual void ConvertToNetworkByteOrder(void * /*writeToHere*/, const void * /*readFromHere*/) const
   {
      MASSERT(false, "This method should never be called.\n");
   }

   virtual void ConvertFromNetworkByteOrder(void * /*writeToHere*/, const void * /*readFromHere*/) const
   {
      MASSERT(false, "This method should never be called.\n");
   }

   Int32DataArray _sizeArray;
   size_t _totalHeldBytes;
   type_code _typeCode;
};

class MessageDataArray : public FixedSizeDataArray<PortableMessageRef>
{
public:
   MessageDataArray() {/* empty */}
   virtual ~MessageDataArray() {/* empty */}

   virtual size_t GetItemSize(int32 index) const 
   {
      const PortableMessage * msg = ItemAt(index).GetItemPointer();
      return msg ? msg->FlattenedSize() : 0;
   }

   virtual type_code TypeCode() const {return B_MESSAGE_TYPE;}

   virtual bool ElementsAreFixedSize() const {return false;}

   virtual void Flatten(uint8 *buffer) const
   {
      size_t writeOffset = 0;
      int numItems = _data.CountItems(); 
      for (int i=0; i<numItems; i++) 
      {
         PortableMessage * next = ItemAt(i).GetItemPointer();
         if (next)
         {
            size_t fs = next->FlattenedSize();
            uint32 writeFs = B_HOST_TO_LENDIAN_INT32(fs);
            WriteData(buffer, &writeOffset, &writeFs, sizeof(writeFs));
            next->Flatten(&buffer[writeOffset]);
            writeOffset += fs;
         }
      }
   }

   // PortableFlattenable interface
   virtual size_t FlattenedSize() const 
   {
      size_t count = 0;
      for (int i=CountItems()-1; i>=0; i--) count += sizeof(uint32) + GetItemSize(i); // 4 bytes for the item size count
      return count;
   }

   virtual status_t Unflatten(const uint8 *buffer, size_t numBytes)
   {
      MakeEmpty();

      size_t readOffset = 0;
      while(readOffset < numBytes)
      {
         uint32 readFs;
         if (ReadData(buffer, numBytes, &readOffset, &readFs, sizeof(readFs)) == false) return B_ERROR;
         readFs = B_LENDIAN_TO_HOST_INT32(readFs);
         if (readOffset + readFs > numBytes) return B_ERROR;  // message size too large for our buffer... corruption?
         PortableMessage * nextMsg = newnothrow PortableMessage;
         if (nextMsg)
         {
            PortableMessageRef msgRef(nextMsg, _defaultMessagePool, _defaultMessageRefPool);
            if (nextMsg->Unflatten(&buffer[readOffset], readFs) != B_NO_ERROR) return B_ERROR;
            if (AddDataItem(&msgRef, sizeof(msgRef)) != B_NO_ERROR) return B_ERROR;
            readOffset += readFs;
         }
         else {WARN_OUT_OF_MEMORY; return B_ERROR;}
      }
      return B_NO_ERROR;
   }

protected:
   virtual AbstractDataArray * Clone() const 
   {
       MessageDataArray * c = newnothrow MessageDataArray;
       if (c == NULL) {WARN_OUT_OF_MEMORY; return NULL;}
       *c = *this;
       return c;
   }

   virtual void PrintToStream()
   {
      int numItems = CountItems();
      for (int i=0; i<numItems; i++)
      {
         const uint8 * dataItem;
         uint32 itemSize = GetItemSize(i);
         if (FindDataItem(i, (const void **)&dataItem) == B_NO_ERROR)
         {
            PortableMessage * msg = ((PortableMessageRef *)dataItem)->GetItemPointer();
            if (msg)
            {
               char buf[5];
               MakePrettyTypeCodeString(msg->what, buf);
               printf("    %i. [what='%s' (%li/0x%lx), flattenedSize=%lu, numFields=%lu]\n", i, buf, msg->what, msg->what, itemSize, msg->CountNames());
            }
         }
         else  printf("    %i. [<Error!>]\n", i);
      }
   }
};

class StringDataArray : public ByteSequenceDataArray
{
public:
   StringDataArray() : ByteSequenceDataArray(B_STRING_TYPE) {/* empty */}
   virtual ~StringDataArray() {/* empty */}

   virtual AbstractDataArray * Clone() const 
   {
       StringDataArray * c = newnothrow StringDataArray;
       if (c == NULL) {WARN_OUT_OF_MEMORY; return NULL;}
       *c = *this;
       return c;
   }

   virtual void PrintToStream()
   {
      PortableString temp;
      int numItems = CountItems();
      for (int i=0; i<numItems; i++)
      {
         const uint8 * dataItem;
         if ((FindDataItem(i, (const void **)&dataItem) == B_NO_ERROR)&&
             (temp.Unflatten(dataItem, GetItemSize(i)) == B_NO_ERROR)) 
               printf("    %i. [%s]\n", i, temp.Cstr());
         else  printf("    %i. [<Error!>]\n", i);
      }
   }
};

/*******************************************
 * PortableMessage class implementation    *
 *******************************************/
 
PortableMessage :: PortableMessage() : what(0), _entries(PortableStringHashFunc)
{
   // empty
}

PortableMessage :: PortableMessage(uint32 w) : what(w), _entries(PortableStringHashFunc)
{
   // empty
}

PortableMessage :: PortableMessage(const PortableMessage &him) : PortableFlattenable(), _entries(PortableStringHashFunc)
{
   *this = him;
}

PortableMessage :: ~PortableMessage() 
{
   MakeEmpty();
}


PortableMessage & PortableMessage :: operator=(const PortableMessage &rhs) 
{
   if (this != &rhs)
   {
      MakeEmpty();
      what = rhs.what;

      // Copy all his entries!
      _eitr it = rhs._entries.GetIterator();
      PortableString * nextKey;
      while((nextKey = it.GetNextKey()) != NULL) 
      {
         AbstractDataArray * clone = (*it.GetNextValue())->Clone();
         if (clone) _entries.Put(*nextKey, clone);
      }
   }
   return *this;
}

status_t PortableMessage :: GetInfo(type_code tr, int32 which, PortableString &name, type_code *typeRet, int32 *count, bool * fixedSize) const 
{
   if ((which < 0)||((size_t) which >= _entries.Size())) return B_ERROR;

   // Find the (i)'th array of the given type...
   int i = -1;
   AbstractDataArray * array = NULL;
   _eitr it = _entries.GetIterator();
   while((i<which)&&(it.GetNextKey(name))) 
   {
      (void) it.GetNextValue(array);
      if ((tr == B_ANY_TYPE)||(tr == array->TypeCode())) i++;
   }
   if (i != which) return B_ERROR;  // nothing found at that index

   if (typeRet)   *typeRet   = array->TypeCode();
   if (count)     *count     = array->CountItems(); 
   if (fixedSize) *fixedSize = array->ElementsAreFixedSize();
   
   return B_NO_ERROR;
}

status_t PortableMessage :: GetInfo(const char *name, type_code *type, int32 *c = 0) const 
{
   return GetInfo(name, type, c, NULL);
}

status_t PortableMessage :: GetInfo(const char *name, type_code *type, bool *fixedSize) const 
{
   return GetInfo(name, type, NULL, fixedSize); 
}

status_t PortableMessage :: GetInfo(const char * name, type_code * type, int32 * c, bool * fixedSize) const
{
   AbstractDataArray * array = GetArray(name, B_ANY_TYPE);
   if (array == NULL) return B_ERROR;
   if (type)       *type     = array->TypeCode();
   if (c)          *c        = array->CountItems();
   if (fixedSize) *fixedSize = array->ElementsAreFixedSize();
   return B_NO_ERROR;
}

int32 PortableMessage :: CountNames(type_code type) const 
{
   if (type == B_ANY_TYPE) return _entries.Size();

   // oops, gotta count just the entries of the given type
   int total = 0;
   _eitr it = _entries.GetIterator();
   AbstractDataArray * nextValue;
   while(it.GetNextValue(nextValue)) if (nextValue->TypeCode() == type) total++;
   return total;
}

bool PortableMessage :: IsEmpty() const 
{
   return (_entries.Size() == 0);
}

void PortableMessage :: PrintToStream() const 
{
   char prettyTypeCodeBuf[5];
   MakePrettyTypeCodeString(what, prettyTypeCodeBuf);

   printf("PortableMessage:  this=%p, what='%s' (%li/0x%lx), entryCount=%li, flatSize=%lu\n", this, prettyTypeCodeBuf, what, what, CountNames(B_ANY_TYPE), (uint32)FlattenedSize());
   _eitr it = _entries.GetIterator();
   PortableString * nextKey;
   while((nextKey = it.GetNextKey()) != NULL) 
   {
       AbstractDataArray * nextValue = *it.GetNextValue();
       type_code tc = nextValue->TypeCode();
       MakePrettyTypeCodeString(tc, prettyTypeCodeBuf);
       printf("  Entry: Name=[%s], CountItems()=%li, TypeCode()='%s' (%li) flatSize=%lu\n", 
            nextKey->Cstr(), nextValue->CountItems(), prettyTypeCodeBuf, tc, (uint32)nextValue->FlattenedSize());
       nextValue->PrintToStream();
   }
}

// Returns an array of the given type, if it exists.  If (etc) is B_ANY_TYPE, then any type array
// is acceptable.
AbstractDataArray * PortableMessage :: GetArray(const PortableString & arrayName, type_code etc) const
{
   AbstractDataArray * array;
   return ((_entries.Get(arrayName, array))&&((etc == B_ANY_TYPE)||(etc == array->TypeCode()))) ? array : NULL;
}

AbstractDataArray * PortableMessage :: GetOrCreateArray(const PortableString & arrayName, type_code etc)
{
   AbstractDataArray * nextEntry = GetArray(arrayName, etc);
   if (nextEntry) return nextEntry;

   // Make sure the problem isn't that there already exists an array, but of the wrong type...
   // If that's the case, we can't create a like-names array of a different type, so fail.
   if (_entries.ContainsKey(arrayName)) return NULL;

   // Oops!  This array doesn't exist; better create it!
   switch(etc)
   {
      // Create appropriate object based on type code
      case B_BOOL_TYPE:    nextEntry = newnothrow BoolDataArray;                         break;
      case B_DOUBLE_TYPE:  nextEntry = newnothrow DoubleDataArray;                       break;
      case B_POINTER_TYPE: nextEntry = newnothrow PointerDataArray;                      break;
      case B_POINT_TYPE:   nextEntry = newnothrow PointDataArray;                        break;
      case B_RECT_TYPE:    nextEntry = newnothrow RectDataArray;                         break;
      case B_FLOAT_TYPE:   nextEntry = newnothrow FloatDataArray;                        break;
      case B_INT64_TYPE:   nextEntry = newnothrow Int64DataArray;                        break;
      case B_INT32_TYPE:   nextEntry = newnothrow Int32DataArray;                        break;
      case B_INT16_TYPE:   nextEntry = newnothrow Int16DataArray;                        break;
      case B_INT8_TYPE:    nextEntry = newnothrow Int8DataArray;                         break;
      case B_MESSAGE_TYPE: nextEntry = newnothrow MessageDataArray;                      break;
      case B_STRING_TYPE:  nextEntry = newnothrow StringDataArray;                       break;
      default:             nextEntry = newnothrow ByteSequenceDataArray(etc);            break;
   }
   if (nextEntry) _entries.Put(arrayName, nextEntry);
             else WARN_OUT_OF_MEMORY;
   return nextEntry;
}


status_t PortableMessage :: Rename(const char *old_entry, const char *new_entry) 
{
   AbstractDataArray * oldArray = GetArray(old_entry, B_ANY_TYPE);
   if (oldArray) 
   { 
      (void) RemoveName(new_entry);                      // destructive rename... delete anybody in our way
      _entries.Remove(PortableString(old_entry));        // remove from under old name
      _entries.Put(PortableString(new_entry), oldArray); // add to under new name
      return B_NO_ERROR;
   }
   else return B_ERROR;
}


size_t PortableMessage :: FlattenedSize() const 
{
   size_t sum = 3 * sizeof(uint32);  // 4 bytes for the protocol revision #, 4 bytes for the number-of-entries field, 4 bytes for what code
   _eitr it = _entries.GetIterator();

   // 4 bytes for the name length, name data, 4 bytes for entry type code, 4 bytes for entry data length, entry data
   PortableString * nextKey;
   while((nextKey = it.GetNextKey()) != NULL) 
   {
      sum += sizeof(uint32) + nextKey->FlattenedSize() + sizeof(uint32) + sizeof(uint32) + (*it.GetNextValue())->FlattenedSize();
   }
   return sum;
}

void PortableMessage :: Flatten(uint8 *buffer) const 
{
   // Format:  0. Protocol revision number (4 bytes, always set to CURRENT_PROTOCOL_VERSION)
   //          1. 'what' code (4 bytes)
   //          2. Number of entries (4 bytes)
   //          3. Entry name length (4 bytes)
   //          4. Entry name string (flattened PortableString)
   //          5. Entry type code (4 bytes)
   //          6. Entry data length (4 bytes)
   //          7. Entry data (n bytes)
   //          8. loop to 3 as necessary

   // Write current protocol version
   size_t writeOffset = 0;
   uint32 networkByteOrder = B_HOST_TO_LENDIAN_INT32(CURRENT_PROTOCOL_VERSION);
   WriteData(buffer, &writeOffset, &networkByteOrder, sizeof(networkByteOrder));

   // Write 'what' code
   networkByteOrder = B_HOST_TO_LENDIAN_INT32(what);
   WriteData(buffer, &writeOffset, &networkByteOrder, sizeof(networkByteOrder));

   // Write number of entries
   networkByteOrder = B_HOST_TO_LENDIAN_INT32(_entries.Size());
   WriteData(buffer, &writeOffset, &networkByteOrder, sizeof(networkByteOrder));

   // Write entries
   _eitr it = _entries.GetIterator();
   PortableString * nextKey;
   while((nextKey = it.GetNextKey()) != NULL)
   {
      // Write entry name length
      size_t keyNameSize = nextKey->FlattenedSize();
      networkByteOrder = B_HOST_TO_LENDIAN_INT32(keyNameSize);
      WriteData(buffer, &writeOffset, &networkByteOrder, sizeof(networkByteOrder));

      // Write entry name
      nextKey->Flatten(&buffer[writeOffset]);
      writeOffset += keyNameSize;

      // Write entry type code
      AbstractDataArray * nextValue = *(it.GetNextValue()); 
      networkByteOrder = B_HOST_TO_LENDIAN_INT32(nextValue->TypeCode());
      WriteData(buffer, &writeOffset, &networkByteOrder, sizeof(networkByteOrder));

      // Write entry data length
      size_t dataSize = nextValue->FlattenedSize();
      networkByteOrder = B_HOST_TO_LENDIAN_INT32(dataSize);
      WriteData(buffer, &writeOffset, &networkByteOrder, sizeof(networkByteOrder));
      
      // Write entry data
      nextValue->Flatten(&buffer[writeOffset]);
      writeOffset += dataSize;
   }
}

status_t PortableMessage :: Unflatten(const uint8 *buffer, size_t inputBufferBytes) 
{
   MakeEmpty();

   size_t readOffset = 0;
   
   // Read and check protocol version number
   size_t networkByteOrder;
   if (ReadData(buffer, inputBufferBytes, &readOffset, &networkByteOrder, sizeof(networkByteOrder)) == false) return B_ERROR;

   uint32 messageProtocolVersion = B_LENDIAN_TO_HOST_INT32(networkByteOrder);
   if ((messageProtocolVersion < OLDEST_SUPPORTED_PROTOCOL_VERSION) ||
       (messageProtocolVersion > CURRENT_PROTOCOL_VERSION)) return B_ERROR;
   
   // Read 'what' code
   if (ReadData(buffer, inputBufferBytes, &readOffset, &networkByteOrder, sizeof(networkByteOrder)) == false) return B_ERROR;
   what = B_LENDIAN_TO_HOST_INT32(networkByteOrder);

   // Read number of entries
   if (ReadData(buffer, inputBufferBytes, &readOffset, &networkByteOrder, sizeof(networkByteOrder)) == false) return B_ERROR;
   uint32 numEntries = B_LENDIAN_TO_HOST_INT32(networkByteOrder);

   // Read entries
   for (size_t i=0; i<numEntries; i++)
   {
      // Read entry name length
      if (ReadData(buffer, inputBufferBytes, &readOffset, &networkByteOrder, sizeof(networkByteOrder)) == false) return B_ERROR;
      size_t nameLength = B_LENDIAN_TO_HOST_INT32(networkByteOrder);

      // Read entry name
      PortableString entryName;
      if (entryName.Unflatten(&buffer[readOffset], nameLength) != B_NO_ERROR) return B_ERROR;
      readOffset += nameLength;

      // Read entry type code
      if (ReadData(buffer, inputBufferBytes, &readOffset, &networkByteOrder, sizeof(networkByteOrder)) == false) return B_ERROR;
      type_code etc = B_LENDIAN_TO_HOST_INT32(networkByteOrder);

      // Read entry data length
      if (ReadData(buffer, inputBufferBytes, &readOffset, &networkByteOrder, sizeof(networkByteOrder)) == false) return B_ERROR;
      size_t eLength = B_LENDIAN_TO_HOST_INT32(networkByteOrder);
   
      AbstractDataArray * nextEntry = GetOrCreateArray(entryName.Cstr(), etc);
      if (nextEntry == NULL) return B_ERROR;

      if (nextEntry->Unflatten(&buffer[readOffset], eLength) != B_NO_ERROR) 
      {
         MakeEmpty();  // fix for occasional crash bug; we were deletint nextEntry here, *and* in the destructor!
         return B_ERROR;
      }
      readOffset += eLength;
   }
   return B_NO_ERROR;
}


status_t PortableMessage :: AddFlatAux(const char * name, const PortableFlattenable & flat, type_code etc, bool prepend)
{
   size_t flatSize = flat.FlattenedSize();
   uint8 * da = newnothrow uint8[flatSize];
   if (da == NULL) {WARN_OUT_OF_MEMORY; return B_ERROR;}
   flat.Flatten(da);
   status_t ret = AddDataAux(name, da, flatSize, etc, prepend);
   if (ret != B_NO_ERROR) delete [] da;
   return ret;
}

status_t PortableMessage :: AddDataAux(const char * name, const uint8 * data, size_t dataSize, type_code etc, bool prepend)
{
   AbstractDataArray * array = GetOrCreateArray(name, etc);
   return array ? (prepend ? array->PrependDataItem(data, dataSize) : array->AddDataItem(data, dataSize)) : B_ERROR;
}

status_t PortableMessage :: AddString(const char *name, const char *val) 
{
   return AddString(name, PortableString(val));
}

status_t PortableMessage :: AddString(const char *name, const PortableString & val) 
{
   return AddFlatAux(name, val, B_STRING_TYPE, false);
}

status_t PortableMessage :: AddInt8(const char *name, int8 val) 
{
   AbstractDataArray * array = GetOrCreateArray(name, B_INT8_TYPE);
   return array ? array->AddDataItem(&val, sizeof(val)) : B_ERROR;
}

status_t PortableMessage :: AddInt16(const char *name, int16 val) 
{
   AbstractDataArray * array = GetOrCreateArray(name, B_INT16_TYPE);
   return array ? array->AddDataItem(&val, sizeof(val)) : B_ERROR;
}

status_t PortableMessage :: AddInt32(const char *name, int32 val) 
{
   AbstractDataArray * array = GetOrCreateArray(name, B_INT32_TYPE);
   return array ? array->AddDataItem(&val, sizeof(val)) : B_ERROR;
}

status_t PortableMessage :: AddInt64(const char *name, int64 val) 
{
   AbstractDataArray * array = GetOrCreateArray(name, B_INT64_TYPE);
   return array ? array->AddDataItem(&val, sizeof(val)) : B_ERROR;
}

status_t PortableMessage :: AddBool(const char *name, bool val) 
{
   AbstractDataArray * array = GetOrCreateArray(name, B_BOOL_TYPE);
   return array ? array->AddDataItem(&val, sizeof(val)) : B_ERROR;
}

status_t PortableMessage :: AddFloat(const char *name, float val) 
{
   AbstractDataArray * array = GetOrCreateArray(name, B_FLOAT_TYPE);
   return array ? array->AddDataItem(&val, sizeof(val)) : B_ERROR;
}

status_t PortableMessage :: AddDouble(const char *name, double val) 
{
   AbstractDataArray * array = GetOrCreateArray(name, B_DOUBLE_TYPE);
   return array ? array->AddDataItem(&val, sizeof(val)) : B_ERROR;
}

status_t PortableMessage :: AddPointer(const char *name, const void * ptr) 
{
   AbstractDataArray * array = GetOrCreateArray(name, B_POINTER_TYPE);
   return array ? array->AddDataItem(&ptr, sizeof(ptr)) : B_ERROR;
}

status_t PortableMessage :: AddPoint(const char *name, const PortablePoint & point) 
{
   AbstractDataArray * array = GetOrCreateArray(name, B_POINT_TYPE);
   return array ? array->AddDataItem(&point, sizeof(point)) : B_ERROR;
}

status_t PortableMessage :: AddRect(const char *name, const PortableRect & rect) 
{
   AbstractDataArray * array = GetOrCreateArray(name, B_RECT_TYPE);
   return array ? array->AddDataItem(&rect, sizeof(rect)) : B_ERROR;
}

status_t PortableMessage :: AddMessage(const char * name, PortableMessageRef ref)
{
   AbstractDataArray * array = GetOrCreateArray(name, B_MESSAGE_TYPE);
   return (array) ? array->AddDataItem(&ref, sizeof(ref)) : B_ERROR;
}

status_t PortableMessage :: AddMessage(const char *name, const PortableMessage & msg) 
{
   AbstractDataArray * array = GetOrCreateArray(name, B_MESSAGE_TYPE);
   if (array)
   {
      PortableMessage * copy = newnothrow PortableMessage(msg);
      if (copy)
      {
         PortableMessageRef ref(copy, _defaultMessagePool, _defaultMessageRefPool);
         return array ? array->AddDataItem(&ref, sizeof(ref)) : B_ERROR;
      }
      else WARN_OUT_OF_MEMORY;
   }
   return B_ERROR;
}

status_t PortableMessage :: AddFlat(const char *name, const PortableFlattenable &obj) 
{
        if (dynamic_cast<const PortableMessage *>(&obj)) return AddMessage(name, (const PortableMessage &)obj);
   else if (dynamic_cast<const PortablePoint *>(&obj))   return AddPoint(name, (const PortablePoint &)obj);
   else if (dynamic_cast<const PortableRect *>(&obj))    return AddRect(name, (const PortableRect &)obj);
   else                                                  return AddFlatAux(name, obj, obj.TypeCode(), false);
}

status_t PortableMessage :: GetElementSize(type_code type, size_t & setSize) const
{
   switch(type)
   {
      case B_BOOL_TYPE:    setSize = sizeof(bool);   break;
      case B_DOUBLE_TYPE:  setSize = sizeof(double); break;
      case B_POINTER_TYPE: setSize = sizeof(void *); break;
      case B_POINT_TYPE:   setSize = sizeof(PortablePoint); break;
      case B_RECT_TYPE:    setSize = sizeof(PortableRect);  break;
      case B_FLOAT_TYPE:   setSize = sizeof(float);  break;
      case B_INT64_TYPE:   setSize = sizeof(int64);  break;
      case B_INT32_TYPE:   setSize = sizeof(int32);  break;
      case B_INT16_TYPE:   setSize = sizeof(int16);  break;
      case B_INT8_TYPE:    setSize = sizeof(int8);   break;
      case B_MESSAGE_TYPE: setSize = sizeof(PortableMessageRef); break;
      default:             setSize = 0;              break;
   }
   return B_NO_ERROR;
}

status_t PortableMessage :: AddData(const char *name, type_code type, const void *data, size_t numBytes) 
{
   return AddData(name, type, data, numBytes, false);
}

status_t PortableMessage :: AddData(const char *name, type_code type, const void *data, size_t numBytes, bool prepend) 
{
   AbstractDataArray * array = GetOrCreateArray(name, type);
   if (array == NULL) return B_ERROR;

   // for primitive types, we do this:
   bool copyData = false;
   size_t elementSize;
   if (GetElementSize(type, elementSize) != B_NO_ERROR) return B_ERROR;
   if (elementSize == 0) 
   {
       // zero indicates a variable-sized data item
       copyData = true;
       elementSize = numBytes;
   }

   if (numBytes % elementSize) return B_ERROR;  // Can't add half an element, silly!
   int numElements = numBytes / elementSize;
   uint8 * dataBuf = (uint8 *) data;
   for (int i=0; i<numElements; i++) 
   {
      uint8 * dataToAdd = dataBuf ? &dataBuf[i*elementSize] : NULL;
      if ((dataToAdd)&&(copyData))
      {
         uint8 * copyFrom = dataToAdd;
         dataToAdd = newnothrow uint8[elementSize];
         if (dataToAdd == NULL) {WARN_OUT_OF_MEMORY; return B_ERROR;}
         memcpy(dataToAdd, copyFrom, elementSize);
      }
      status_t ret = prepend ? array->PrependDataItem(dataToAdd, elementSize)
                             : array->AddDataItem(dataToAdd, elementSize);
      if (ret != B_NO_ERROR)
      {
         if (copyData) delete [] dataToAdd;
         return ret;
      }
   }
   return B_NO_ERROR;
}


status_t PortableMessage :: RemoveData(const char *name, int32 index) 
{
   AbstractDataArray * array = GetArray(name, B_ANY_TYPE);
   if (array != NULL) 
   {
      status_t ret = array->RemoveDataItem(index);
      if (array->CountItems() == 0) return RemoveName(name);
      return ret;
   }
   else return B_ERROR;
}

status_t PortableMessage :: RemoveName(const char *name) 
{
   AbstractDataArray * array;
   if (_entries.Remove(PortableString(name), array)) 
   {
      delete array;
      return B_NO_ERROR;
   }
   else return B_ERROR;
}

void PortableMessage :: MakeEmpty()
{
   // Delete all the arrays before clearing the Hashtable!
   _eitr it = _entries.GetIterator();
   AbstractDataArray * nextArray;
   while(it.GetNextValue(nextArray)) delete nextArray;
   _entries.Clear();
}      


status_t PortableMessage :: FindString(const char *name, int32 index, const char ** setMe) const 
{
   size_t junk;
   return FindDataAux(name, index, (const uint8 **)setMe, &junk, B_STRING_TYPE);
}

status_t PortableMessage :: FindString(const char *name, int32 index, PortableString & str) const 
{
   return FindFlatAux(name, index, str);
}

status_t PortableMessage :: FindFlatAux(const char * name, int32 index, PortableFlattenable & flat) const
{
   AbstractDataArray * ada = GetArray(name, B_ANY_TYPE);
   if ((ada)&&(flat.AllowsTypeCode(ada->TypeCode())))
   {
      const uint8 * data;
      status_t ret = ada->FindDataItem(index, (const void **) &data);
      if (ret == B_NO_ERROR) return flat.Unflatten(data, ada->GetItemSize(index));
   }
   return B_ERROR;
}

status_t PortableMessage :: FindDataAux(const char * name, int32 index, const uint8 ** setDataLoc, size_t * setSize, type_code tc) const
{
   AbstractDataArray * array = GetArray(name, tc);
   if (array)
   {
      status_t ret = array->FindDataItem(index, (const void **) setDataLoc);
      if (ret != B_NO_ERROR) return ret;
      if (setSize) *setSize = array->GetItemSize(index);
      return ret;
   }
   else return B_ERROR;
}

status_t PortableMessage :: FindDataItemAux(const char * name, int32 index, type_code tc, void * setValue, size_t valueSize) const
{
   AbstractDataArray * array = GetArray(name, tc);
   if (array == NULL) return B_ERROR;
   const void * addressOfValue;
   status_t ret = array->FindDataItem(index, &addressOfValue);
   if (ret != B_NO_ERROR) return ret;
   memcpy(setValue, addressOfValue, valueSize);
   return B_NO_ERROR;
}

status_t PortableMessage :: FindInt8(const char *name, int32 index, int8 *val) const 
{
   return FindDataItemAux(name, index, B_INT8_TYPE, val, sizeof(int8));
}

status_t PortableMessage :: FindInt16(const char *name, int32 index, int16 *val) const 
{
   return FindDataItemAux(name, index, B_INT16_TYPE, val, sizeof(int16));
}

status_t PortableMessage :: FindInt32(const char *name, int32 index, int32 *val) const 
{
   return FindDataItemAux(name, index, B_INT32_TYPE, val, sizeof(int32));
}

status_t PortableMessage :: FindInt64(const char *name, int32 index, int64 *val) const 
{
   return FindDataItemAux(name, index, B_INT64_TYPE, val, sizeof(int64));
}

status_t PortableMessage :: FindBool(const char *name, int32 index, bool *val) const 
{
   return FindDataItemAux(name, index, B_BOOL_TYPE, val, sizeof(bool));
}

status_t PortableMessage :: FindFloat(const char *name, int32 index, float *val) const 
{
   return FindDataItemAux(name, index, B_FLOAT_TYPE, val, sizeof(float));
}

status_t PortableMessage :: FindDouble(const char *name, int32 index, double * val) const 
{
   return FindDataItemAux(name, index, B_DOUBLE_TYPE, val, sizeof(double));
}

status_t PortableMessage :: FindPointer(const char *name, int32 index, void ** ptr) const 
{
   return FindDataItemAux(name, index, B_POINTER_TYPE, ptr, sizeof(void *));
}

status_t PortableMessage :: FindPoint(const char *name, int32 index, PortablePoint & point) const 
{
   AbstractDataArray * array = GetArray(name, B_POINT_TYPE);
   if (array == NULL) return B_ERROR;
   const void * addressOfValue;
   status_t ret = array->FindDataItem(index, &addressOfValue);
   if (ret != B_NO_ERROR) return ret;
   point = *((PortablePoint *)addressOfValue);
   return B_NO_ERROR;
}

status_t PortableMessage :: FindRect(const char *name, int32 index, PortableRect & rect) const 
{
   AbstractDataArray * array = GetArray(name, B_RECT_TYPE);
   if (array == NULL) return B_ERROR;
   const void * addressOfValue;
   status_t ret = array->FindDataItem(index, &addressOfValue);
   if (ret != B_NO_ERROR) return ret;
   rect = *((PortableRect *)addressOfValue);
   return B_NO_ERROR;
}

status_t PortableMessage :: FindMessage(const char *name, int32 index, PortableMessage &msg) const 
{
   PortableMessageRef msgRef;
   if (FindMessage(name, index, msgRef) == B_NO_ERROR)
   {
      const PortableMessage * m = msgRef.GetItemPointer();
      if (m) 
      {
         msg = *m;
         return B_NO_ERROR;
      }
   }
   return B_ERROR;
}

status_t PortableMessage :: FindMessage(const char * name, int32 index, PortableMessageRef & ref) const
{
   AbstractDataArray * array = GetArray(name, B_MESSAGE_TYPE);
   if (array == NULL) return B_ERROR;
   const void * addressOfValue;
   status_t ret = array->FindDataItem(index, &addressOfValue);
   if (ret != B_NO_ERROR) return B_ERROR;
   ref = *((PortableMessageRef *)addressOfValue);
   return B_NO_ERROR;
}

status_t PortableMessage :: FindFlat(const char *name, int32 index, PortableFlattenable &obj) const
{
        if (dynamic_cast<PortableMessage *>(&obj)) return FindMessage(name, index, (PortableMessage &)obj);
   else if (dynamic_cast<PortablePoint *>(&obj))  return FindPoint(name, index, (PortablePoint &)obj);
   else if (dynamic_cast<PortableRect *>(&obj))   return FindRect(name, index, (PortableRect &)obj);
   else                                           return FindFlatAux(name, index, obj);
}

status_t PortableMessage :: FindData(const char *name, type_code etc, int32 index, const void **data, size_t *numBytes) const
{
   void * returnData;
   status_t ret = FindDataPointer(name, etc, index, &returnData, numBytes);
   if (ret == B_NO_ERROR) *data = returnData;
   return ret;
}

status_t PortableMessage :: FindDataPointer(const char *name, type_code etc, int32 index, void **data, size_t *numBytes) const 
{
   const uint8 * dataLoc;
   status_t ret = FindDataAux(name, index, &dataLoc, numBytes, etc);
   if (ret == B_NO_ERROR) *data = (void *) dataLoc;  // breaks const correctness, but oh well
   return ret;
}

status_t PortableMessage :: ReplaceString(const char *name, int32 index, const PortableString & string) 
{
   return ReplaceFlatItemAux(name, index, string, B_STRING_TYPE);
}

status_t PortableMessage :: ReplaceFlatItemAux(const char * name, int32 index, const PortableFlattenable & flat, type_code tc)
{
   // Create flattened data buffer
   size_t flatSize = flat.FlattenedSize();
   uint8 * newBuf = newnothrow uint8[flatSize];
   if (newBuf == NULL) {WARN_OUT_OF_MEMORY; return B_ERROR;}
   flat.Flatten(newBuf);
   status_t ret = ReplaceDataAux(name, index, newBuf, flatSize, tc);
   if (ret != B_NO_ERROR) delete [] newBuf;
   return ret;
}

status_t PortableMessage :: ReplaceDataAux(const char * name, int32 index, uint8 * dataBuf, size_t bufSize, type_code tc)
{
   AbstractDataArray * array = GetArray(name, tc);
   return array ? array->ReplaceDataItem(index, dataBuf, bufSize) : B_ERROR;
}

status_t PortableMessage :: ReplaceInt8(const char *name, int32 index, int8 val) 
{
   AbstractDataArray * array = GetArray(name, B_INT8_TYPE);
   return array ? array->ReplaceDataItem(index, &val, sizeof(val)) : B_ERROR;
}

status_t PortableMessage :: ReplaceInt16(const char *name, int32 index, int16 val) 
{
   AbstractDataArray * array = GetArray(name, B_INT16_TYPE);
   return array ? array->ReplaceDataItem(index, &val, sizeof(val)) : B_ERROR;
}

status_t PortableMessage :: ReplaceInt32(const char *name, int32 index, int32 val) 
{
   AbstractDataArray * array = GetArray(name, B_INT32_TYPE);
   return array ? array->ReplaceDataItem(index, &val, sizeof(val)) : B_ERROR;
}

status_t PortableMessage :: ReplaceInt64(const char *name, int32 index, int64 val) 
{
   AbstractDataArray * array = GetArray(name, B_INT64_TYPE);
   return array ? array->ReplaceDataItem(index, &val, sizeof(val)) : B_ERROR;
}

status_t PortableMessage :: ReplaceBool(const char *name, int32 index, bool val) 
{
   AbstractDataArray * array = GetArray(name, B_BOOL_TYPE);
   return array ? array->ReplaceDataItem(index, &val, sizeof(val)) : B_ERROR;
}

status_t PortableMessage :: ReplaceFloat(const char *name, int32 index, float val) 
{
   AbstractDataArray * array = GetArray(name, B_FLOAT_TYPE);
   return array ? array->ReplaceDataItem(index, &val, sizeof(val)) : B_ERROR;
}

status_t PortableMessage :: ReplaceDouble(const char *name, int32 index, double val) 
{
   AbstractDataArray * array = GetArray(name, B_DOUBLE_TYPE);
   return array ? array->ReplaceDataItem(index, &val, sizeof(val)) : B_ERROR;
}

status_t PortableMessage :: ReplacePointer(const char *name, int32 index, const void * ptr) 
{
   AbstractDataArray * array = GetArray(name, B_POINTER_TYPE);
   return array ? array->ReplaceDataItem(index, &ptr, sizeof(ptr)) : B_ERROR;
}

status_t PortableMessage :: ReplacePoint(const char *name, int32 index, const PortablePoint &point) 
{
   AbstractDataArray * array = GetArray(name, B_POINT_TYPE);
   return array ? array->ReplaceDataItem(index, &point, sizeof(point)) : B_ERROR;
}

status_t PortableMessage :: ReplaceRect(const char *name, int32 index, const PortableRect &rect) 
{
   AbstractDataArray * array = GetArray(name, B_RECT_TYPE);
   return array ? array->ReplaceDataItem(index, &rect, sizeof(rect)) : B_ERROR;
}

status_t PortableMessage :: ReplaceMessage(const char *name, int32 index, const PortableMessage &msg)
{
   PortableMessage * copy = newnothrow PortableMessage(msg);
   if (copy)
   {
      PortableMessageRef msgRef(copy, _defaultMessagePool, _defaultMessageRefPool);
      return ReplaceMessage(name, index, msgRef);
   }
   else WARN_OUT_OF_MEMORY;
   return B_ERROR;
}

status_t PortableMessage :: ReplaceMessage(const char * name, int32 index, PortableMessageRef msgRef)
{
   AbstractDataArray * array = GetArray(name, B_MESSAGE_TYPE);
   if (array) return array->ReplaceDataItem(index, &msgRef, sizeof(msgRef));
   return B_ERROR;
}

status_t PortableMessage :: ReplaceFlat(const char *name, int32 index, const PortableFlattenable &obj) 
{
        if (dynamic_cast<const PortableMessage *>(&obj)) return ReplaceMessage(name, index, (const PortableMessage &)obj);
   else if (dynamic_cast<const PortablePoint *>(&obj))   return ReplacePoint(name, index, (const PortablePoint &)obj);
   else if (dynamic_cast<const PortableRect *>(&obj))    return ReplaceRect(name, index, (const PortableRect &)obj);
   else                                                  return ReplaceFlatItemAux(name, index, obj, obj.TypeCode());
}

status_t PortableMessage :: ReplaceData(const char *name, type_code type, int32 index, const void *data, size_t numBytes) 
{
   AbstractDataArray * array = GetArray(name, type);
   if (array == NULL) return B_ERROR;
   
   // for primitive types, we do this:
   bool copyData = false;
   size_t elementSize;
   if (GetElementSize(type, elementSize) != B_NO_ERROR) return B_ERROR;
   if (elementSize == 0) 
   {
       // zero indicates a variable-sized data item
       copyData = true;
       elementSize = numBytes;
   }

   if (numBytes % elementSize) return B_ERROR;  // Can't add half an element, silly!
   int numElements = numBytes / elementSize;
   uint8 * dataBuf = (uint8 *) data;
   for (int i=index; i<index+numElements; i++) 
   {
      uint8 * dataToAdd = &dataBuf[i*elementSize];
      if (copyData)
      {
         uint8 * copyFrom = dataToAdd;
         dataToAdd = newnothrow uint8[elementSize];
         if (dataToAdd == NULL) {WARN_OUT_OF_MEMORY; return B_ERROR;}
         memcpy(dataToAdd, copyFrom, elementSize);
      }
      status_t ret = array->ReplaceDataItem(i, dataToAdd, elementSize);
      if (ret != B_NO_ERROR)
      {
         if (copyData) delete [] dataToAdd;
         return ret;
      }
   }
   return B_NO_ERROR;
}

status_t PortableMessage :: MoveName(const char * name, PortableMessage &moveTo)
{
   AbstractDataArray * array = GetArray(name, B_ANY_TYPE);
   if (array)
   {
      (void) moveTo.RemoveName(name);  // clear out any previous data he had under that name
      PortableString pm(name);
      _entries.Remove(pm);
      moveTo._entries.Put(pm, array);
      return B_NO_ERROR;
   }
   else return B_ERROR;
}

bool PortableMessage :: HasName(const char * name, type_code type) const
{
   return (GetArray(name, type) != NULL);
}

status_t PortableMessage :: CopyName(const char * name, PortableMessage &copyTo) const
{
   AbstractDataArray * array = GetArray(name, B_ANY_TYPE);
   if (array)
   {
      AbstractDataArray * clone = array->Clone();
      if (clone)
      {
         (void) copyTo.RemoveName(name);  // clear out any previous data he had under that name
         copyTo._entries.Put(PortableString(name), clone);
         return B_NO_ERROR;
      }
   }
   return B_ERROR;
}

PortableMessageFieldNameIterator PortableMessage :: GetFieldNameIterator(type_code type) const
{
   _eitr * clone = newnothrow _eitr(_entries.GetIterator());
   if (clone == NULL) WARN_OUT_OF_MEMORY;
   return PortableMessageFieldNameIterator(*this, clone, type);
}

const char * PortableMessage :: PeekNextFieldName(_eitr * it, type_code tc) const
{
   if (tc == B_ANY_TYPE) 
   {
      PortableString * ret = it->PeekNextKey();
      return ret ? ret->Cstr() : NULL;
   }
   else
   {
      const char * ret = NULL;

      // Read until the next array to get is of type (tc)...
      AbstractDataArray * array;
      while(it->PeekNextValue(array))
      {
         if (array->TypeCode() == tc) return ret;

         // iterate along...
         ret = it->GetNextKey()->Cstr();
         (void)it->GetNextValue();
      }
      return NULL;
   }
}

const char * PortableMessage :: GetNextFieldName(_eitr * it, type_code tc) const
{
   if (tc == B_ANY_TYPE) 
   {
      PortableString * ret = it->GetNextKey();
      return ret ? ret->Cstr() : NULL;
   }
   else
   {
      // Read until the array is of type (tc)...
      AbstractDataArray * array;
      while(it->GetNextValue(array))
      {
         PortableString * ret = it->GetNextKey();
         if (array->TypeCode() == tc) return (ret ? ret->Cstr() : NULL);
      }
   }
   return NULL;
}

void PortableMessage :: SetInternalDefaultPools(PortableMessageRef::ItemPool * ip, PortableMessageRef::RefPool * rp)
{
   _defaultMessagePool = ip;
   _defaultMessageRefPool = rp;
}

status_t PortableMessage :: PrependString(const char *name, const char *val) 
{
   return PrependString(name, PortableString(val));
}

status_t PortableMessage :: PrependString(const char *name, const PortableString & val) 
{
   return AddFlatAux(name, val, B_STRING_TYPE, true);
}

status_t PortableMessage :: PrependInt8(const char *name, int8 val) 
{
   AbstractDataArray * array = GetOrCreateArray(name, B_INT8_TYPE);
   return array ? array->PrependDataItem(&val, sizeof(val)) : B_ERROR;
}

status_t PortableMessage :: PrependInt16(const char *name, int16 val) 
{
   AbstractDataArray * array = GetOrCreateArray(name, B_INT16_TYPE);
   return array ? array->PrependDataItem(&val, sizeof(val)) : B_ERROR;
}

status_t PortableMessage :: PrependInt32(const char *name, int32 val) 
{
   AbstractDataArray * array = GetOrCreateArray(name, B_INT32_TYPE);
   return array ? array->PrependDataItem(&val, sizeof(val)) : B_ERROR;
}

status_t PortableMessage :: PrependInt64(const char *name, int64 val) 
{
   AbstractDataArray * array = GetOrCreateArray(name, B_INT64_TYPE);
   return array ? array->PrependDataItem(&val, sizeof(val)) : B_ERROR;
}

status_t PortableMessage :: PrependBool(const char *name, bool val) 
{
   AbstractDataArray * array = GetOrCreateArray(name, B_BOOL_TYPE);
   return array ? array->PrependDataItem(&val, sizeof(val)) : B_ERROR;
}

status_t PortableMessage :: PrependFloat(const char *name, float val) 
{
   AbstractDataArray * array = GetOrCreateArray(name, B_FLOAT_TYPE);
   return array ? array->PrependDataItem(&val, sizeof(val)) : B_ERROR;
}

status_t PortableMessage :: PrependDouble(const char *name, double val) 
{
   AbstractDataArray * array = GetOrCreateArray(name, B_DOUBLE_TYPE);
   return array ? array->PrependDataItem(&val, sizeof(val)) : B_ERROR;
}

status_t PortableMessage :: PrependPointer(const char *name, const void * ptr) 
{
   AbstractDataArray * array = GetOrCreateArray(name, B_POINTER_TYPE);
   return array ? array->PrependDataItem(&ptr, sizeof(ptr)) : B_ERROR;
}

status_t PortableMessage :: PrependPoint(const char *name, const PortablePoint & point) 
{
   AbstractDataArray * array = GetOrCreateArray(name, B_POINT_TYPE);
   return array ? array->PrependDataItem(&point, sizeof(point)) : B_ERROR;
}

status_t PortableMessage :: PrependRect(const char *name, const PortableRect & rect) 
{
   AbstractDataArray * array = GetOrCreateArray(name, B_RECT_TYPE);
   return array ? array->PrependDataItem(&rect, sizeof(rect)) : B_ERROR;
}

status_t PortableMessage :: PrependMessage(const char * name, PortableMessageRef ref)
{
   AbstractDataArray * array = GetOrCreateArray(name, B_MESSAGE_TYPE);
   return (array) ? array->PrependDataItem(&ref, sizeof(ref)) : B_ERROR;
}

status_t PortableMessage :: PrependMessage(const char *name, const PortableMessage & msg) 
{
   AbstractDataArray * array = GetOrCreateArray(name, B_MESSAGE_TYPE);
   if (array)
   {
      PortableMessage * copy = newnothrow PortableMessage(msg);
      if (copy)
      {
         PortableMessageRef ref(copy, _defaultMessagePool, _defaultMessageRefPool);
         return array ? array->PrependDataItem(&ref, sizeof(ref)) : B_ERROR;
      }
      else WARN_OUT_OF_MEMORY;
   }
   return B_ERROR;
}

status_t PortableMessage :: PrependFlat(const char *name, const PortableFlattenable &obj) 
{
        if (dynamic_cast<const PortableMessage *>(&obj)) return PrependMessage(name, (const PortableMessage &)obj);
   else if (dynamic_cast<const PortablePoint *>(&obj))   return PrependPoint(name, (const PortablePoint &)obj);
   else if (dynamic_cast<const PortableRect *>(&obj))    return PrependRect(name, (const PortableRect &)obj);
   else                                                  return AddFlatAux(name, obj, obj.TypeCode(), true);
}

status_t PortableMessage :: PrependData(const char *name, type_code type, const void *data, size_t numBytes) 
{
   return AddData(name, type, data, numBytes, true);
}

