/* This file is Copyright 2001 Level Control Systems.  See the included LICENSE.txt file for details. */

/******************************************************************************
/
/   File:       PortableMessage.h
/
/   Description:   OS-independent version of Be's BMessage class.
/         Doesn't contain all of BMessage's functionality,
/         but keeps enough of it to be useful as a portable
/         data object (dictionary) in cross-platform apps.
/
*******************************************************************************/

#ifndef _PORTABLEMESSAGE_H
#define _PORTABLEMESSAGE_H

#include "refcount/PortableRefCount.h"
#include "support/PortableTypeDefs.h"
#include "support/PortableFlattenable.h"
#include "support/PortablePoint.h"
#include "support/PortableRect.h"
#include "string/PortableString.h"
#include "hashtable/PortableHashtable.h"
#include "refcount/PortableRefCount.h"

class PortableMessage;
typedef PortableRef<PortableMessage> PortableMessageRef;

class PortableMessage;

/** This function returns a pointer to a singleton ObjectPool that can be 
 *  used to minimize the number of PortableMessage allocations and deletions by
 *  recycling the PortableMessage objects.  Note that the returned ObjectPool is
 *  NOT THREAD SAFE, so you probably shouldn't use it in multithreaded
 *  programs.
 */
PortableMessageRef::ItemPool * GetMessagePool();

/** This function returns a pointer to a singleton ObjectPool that can be 
 *  used to minimize the number of PortableMessageRef allocations and deletions by
 *  recycling the PortableMessageRef objects.  Note that the returned ObjectPool is
 *  NOT THREAD SAFE, so you probably shouldn't use it in multithreaded
 *  programs.
 */
PortableMessageRef::RefPool * GetMessageRefPool();

/** Convenience method:  Gets a PortableMessage from the default PortableMessage pool and returns it. 
 *  @param what The 'what' code to set in the returned PortableMessage.
 *  @return A PortableMessage object (caller assumes ownership), or NULL on failure (out of memory).
 */
PortableMessage * GetMessageFromPool(uint32 what = 0L);

/** Convenience method:  Gets a PortableMessage from the message pool, makes it equal to (copyMe), and returns it. 
 *  @param copyMe A PortableMessage to return a clone of.
 *  @return A PortableMessage object (caller assumes ownership), or NULL on failure (out of memory).
 */
PortableMessage * GetMessageFromPool(const PortableMessage & copyMe);

// these declarations are for internal use only
class AbstractDataArray;
typedef PortableHashtableIterator<PortableString, AbstractDataArray *> _eitr;   

/** This is an iterator that allows you to efficiently iterate over the field names in a PortableMessage. 
 *  It can be allocated only by the PortableMessage::GetFieldNameIterator() method.
 */
class PortableMessageFieldNameIterator
{
public:
   /** Default constructor.   
    *  Creates an "empty" iterator;  Use the assignment operator to turn the iterator into something useful.
    */
   PortableMessageFieldNameIterator();

   /** Copy Constructor. */
   PortableMessageFieldNameIterator(const PortableMessageFieldNameIterator & copyMe);

   /** Assignment operator. */
   PortableMessageFieldNameIterator & operator = (const PortableMessageFieldNameIterator & rhs);

   /** Destructor. */
   ~PortableMessageFieldNameIterator();

   /** Returns true iff there are more field names available to be traversed.  */
   bool HasMoreFieldNames() const;

   /**
    *  Places the next field name into (name) and bumps the iterator position.
    *  @param name On success, the next field name is written into this object.
    *  @return true iff successful, false if there were no more names are left to get.
    */
   bool GetNextFieldName(PortableString & name);

   /**
    *  Returns the next field name and bumps the iterator position, or returns NULL 
    *  if there are no more field names left to get.  (Note that
    *  the returned string may become invalid when the hashtable is changed)
    */
   const char * GetNextFieldName();

   /** 
    *  Places the next field name into (name) without modifying the state of the traversal.
    *  @param name On success, the next field name is written into this object.
    *  @return True on success, or false if there were no more field names left to traverse.
    */
   bool PeekNextFieldName(PortableString & name) const;

   /** 
    *  Returns The next field name on success, or NULL if there were no more field
    *  names left to traverse.  Does not modify the state of the traversal.
    */
   const char * PeekNextFieldName() const;

private:
   friend class PortableMessage;

   PortableMessageFieldNameIterator(const PortableMessage & owner, _eitr * iter, type_code tc)
    : _owner(&owner), _iter(iter), _typeCode(tc) {/* empty */}

   const PortableMessage * _owner;
   _eitr * _iter;
   type_code _typeCode;
};

// Version number of the PortableMessage serialization protocol.
// Will be the first four bytes of each serialized PortableMessage buffer.
// Only buffers beginning with version numbers between these two 
// constants (inclusive) will be Unflattened as PortableMessages.
#define OLDEST_SUPPORTED_PROTOCOL_VERSION 1347235888 // 'PM00'
#define CURRENT_PROTOCOL_VERSION          1347235888 // 'PM00'

/** 
 *  This class is similar to a BMessage, but portable.  
 *  It only acts as a serializable data container; it does not have
 *  include any threading capabilities.  Unless otherwise noted, all
 *  methods behave similarly to their BMessage counterparts.  (Exception:
 *  the only error codes returned are B_ERROR and B_NO_ERROR)
 */
class PortableMessage : public PortableFlattenable 
{
public:
   /** 32 bit what code, for quick identification of message types.  Set this however you like. */
   uint32   what;

   /** Default Constructor. */
   PortableMessage();

   /** Constructor.
    *  @param what The 'what' member variable will be set to the value you specify here.
    */
   PortableMessage(uint32 what);

   /** Copy Constructor. */
   PortableMessage(const PortableMessage &a_message);

   /** Destructor. */
   virtual ~PortableMessage();

   /** Assignment operator. */
   PortableMessage &operator=(const PortableMessage &msg);

   /** Be-style field info query method. 
    *  @param typeRequested The B_*_TYPE you are interested in limiting your query to.  (Use B_ANY_TYPE for a general query)
    *  @param which The index of the field in the message (if typeRequested was B_ANY_TYPE), or amongst fields of the given type (if typeRequested was some other type). 
    *  @param setName On success, the name of the found field is copied into this object.
    *  @param typeReturned On success, the type code of the found field is copied into this object.
    *  @param count On success, the number of items stored in the found field is copied into this object.
    *  @param fixed_size On success, *fixed_size is set to reflect whether the returned field's objects are all the same size, or not.
    *  @return B_NO_ERROR on success, or B_ERROR if the requested field couldn't be found.
    */
   status_t GetInfo(type_code typeRequested, int32 which, PortableString & setName, type_code *typeReturned, int32 *count = NULL, bool * fixed_size = NULL) const;

   /** Be-style field info query method
    *  @param name Name of the field type look for.
    *  @param type On sucess, the type code of the found field is copied into this object.
    *  @param c On success, the number of items in the found field is copied into this object.
    *  @return B_NO_ERROR on success, or B_ERROR if the requested field couldn't be found.
    */
   status_t GetInfo(const char *name, type_code *type, int32 *c) const;

   /** Be-style field info query method
    *  @param name Name of the field type look for.
    *  @param type On sucess, the type code of the found field is copied into this object.
    *  @param fixed_size On success, *fixed_size is set to reflect whether the returned field's objects are all the same size, or not.
    *  @return B_NO_ERROR on success, or B_ERROR if the requested field couldn't be found.
    */
   status_t GetInfo(const char *name, type_code *type, bool *fixed_size) const;

   /** Be-style field info query method
    *  @param name Name of the field type look for.
    *  @param type On sucess, the type code of the found field is copied into this object.
    *  @param c On success, the number of items in the found field is copied into this object.
    *  @param fixed_size On success, *fixed_size is set to reflect whether the returned field's objects are all the same size, or not.
    *  @return B_NO_ERROR on success, or B_ERROR if the requested field couldn't be found.
    */
   status_t GetInfo(const char *name, type_code *type, int32 *c, bool *fixed_size) const;

   /** Returns the number of field names of the given type that are present in the PortableMessage.
    *  @param type The type of field to count, or B_ANY_TYPE to count all field types.
    *  @return The number of matching fields
    */
   int32 CountNames(type_code type = B_ANY_TYPE) const;

   /** @return true iff there are no fields in this PortableMessage. */
   bool IsEmpty() const;

   /** Prints debug info describing the contents of this PortableMessage to stdout. */
   void PrintToStream() const;

   /** Renames a field.
    *  @param old_entry Field name to rename from.
    *  @param new_entry Field name to rename to.  If a field with this name already exists,
    *                   it will be replaced.
    *  @return B_NO_ERROR on success, or B_ERROR if a field named (old_entry) couldn't be found. 
    */
   status_t Rename(const char *old_entry, const char *new_entry);

   /** Returns false (PortableMessages can be of various sizes, depending on how many fields they have, etc.) */
   virtual bool IsFixedSize() const {return false;}

   /** Returns B_MESSAGE_TYPE */
   virtual type_code TypeCode() const {return B_MESSAGE_TYPE;}

   /** Returns The number of bytes it would take to flatten this PortableMessage into a byte buffer. */
   virtual size_t FlattenedSize() const;
 
   /**
    *  Converts this PortableMessage into a flattened buffer of bytes that can be saved to disk
    *  or sent over a network, and later converted back into an identical PortableMessage object.
    *  @param buffer The byte buffer to store the PortableMessage into.  It must be FlattenedSize() bytes long.
    */
   virtual void Flatten(uint8 *buffer) const;

   /**
    *  Convert the given byte buffer back into a PortableMessage.  Any previous contents of
    *  this PortableMessage will be erased, and replaced with the data specified in the byte buffer.
    *  @param buf Pointer to a byte buffer containing a flattened PortableMessage to restore.
    *  @size The number of bytes in the flattened byte buffer.
    *  @return B_NO_ERROR if the buffer was successfully Unflattened, or B_ERROR if there
    *          was an error (usually meaning the buffer was corrupt, or out-of-memory)
    */
   virtual status_t Unflatten(const uint8 *buf, size_t size);

   /** Adds a new string to the PortableMessage.
    *  @param name Name of the field to add (or add to)
    *  @param a_string The string to add 
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t AddString(const char *name, const char *a_string);

   /** Adds a new string to the PortableMessage.
    *  @param name Name of the field to add (or add to)
    *  @param a_string The string to add
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t AddString(const char *name, const PortableString & a_string);

   /** Adds a new int8 to the PortableMessage.
    *  @param name Name of the field to add (or add to)
    *  @param val The int8 to add 
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t AddInt8(const char *name, int8 val);

   /** Adds a new int16 to the PortableMessage.
    *  @param name Name of the field to add (or add to)
    *  @param val The int16 to add
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t AddInt16(const char *name, int16 val);

   /** Adds a new int32 to the PortableMessage.
    *  @param name Name of the field to add (or add to)
    *  @param val The int32 to add
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t AddInt32(const char *name, int32 val);

   /** Adds a new int64 to the PortableMessage.
    *  @param name Name of the field to add (or add to)
    *  @param val The int64 to add
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t AddInt64(const char *name, int64 val);

   /** Adds a new boolean to the PortableMessage.
    *  @param name Name of the field to add (or add to)
    *  @param a_boolean The boolean to add
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t AddBool(const char *name, bool a_boolean);

   /** Adds a new float to the PortableMessage.
    *  @param name Name of the field to add (or add to)
    *  @param a_float The float to add
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t AddFloat(const char *name, float a_float);

   /** Adds a new double to the PortableMessage.
    *  @param name Name of the field to add (or add to)
    *  @param a_double The double to add
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t AddDouble(const char *name, double a_double);

   /** Adds a new PortableMessage to the PortableMessage.
    *  Note that this method is less efficient than the
    *  AddMessage(PortableMessageRef) method, as this method
    *  necessitates copying all the data in (msg).
    *  @param name Name of the field to add (or add to)
    *  @param msg The PortableMessage to add
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t AddMessage(const char *name, const PortableMessage & msg);

   /** Adds a new PortableMessage to the PortableMessage.
    *  Note that this method is more efficient than the
    *  AddMessage(const PortableMessage &) method, as this method
    *  need not make a copy of the referenced PortableMessage.
    *  @param name Name of the field to add (or add to)
    *  @param msgRef Reference to the PortableMessage to add
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t AddMessage(const char *name, PortableMessageRef msgRef); 

   /** Adds a new pointer value to the PortableMessage.
    *  @param name Name of the field to add (or add to)
    *  @param ptr The pointer value to add
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t AddPointer(const char *name, const void * ptr);

   /** Adds a new point to the PortableMessage.
    *  @param name Name of the field to add (or add to)
    *  @param point The point to add
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t AddPoint(const char *name, const PortablePoint & point);

   /** Adds a new rect to the PortableMessage.
    *  @param name Name of the field to add (or add to)
    *  @param rect The rect to add
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t AddRect(const char *name, const PortableRect & rect);

   /** Adds a new PortableFlattenable object to the PortableMessage.
    *  @param name Name of the field to add (or add to)
    *  @param obj The PortableFlattenable object to add
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t AddFlat(const char *name, const PortableFlattenable &obj);

   /** Generic method, capable of adding any type of data to the PortableMessage.
    *  @param name Name of the field to add (or add to)
    *  @param type The B_*_TYPE code indicating the type of data that will be added.
    *  @param data A pointer to the bytes of data to add.  The data is copied out of
    *              this byte buffer, and exactly how it is interpreted depends on (type).
    *              (data) may be NULL, in which case default objects (or uninitialized byte space) is added.
    *  Note:  If you are adding B_MESSAGE_TYPE, then (data) must point to a PortableMessageRef
    *         object, and NOT a PortableMessage object, or flattened-PortableMessage-buffer!
    *         (This is inconsistent with BMessage::AddData, which expects a flattened BMessage buffer.  Sorry!)
    *  @param numBytes The number of bytes that are pointed to by (data).  If (type)
    *                  is a known type such as B_INT32_TYPE or B_RECT_TYPE, (numBytes)
    *                  may be a multiple of the datatype's size, allowing you to add
    *                  multiple entries with a single call.
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or (numBytes) isn't a multiple (type)'s 
    *          known size, or a type conflict occurred.
    */
   status_t AddData(const char *name, type_code type, const void *data, size_t numBytes);

   /** Prepends a new string to the beginning of a field array in the PortableMessage.
    *  @param name Name of the field to add (or prepend to)
    *  @param a_string The string to add or prepend
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t PrependString(const char *name, const char *a_string);

   /** Prepends a new string to the beginning of a field array in the PortableMessage.
    *  @param name Name of the field to add (or prepend to)
    *  @param a_string The string to add or prepend
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t PrependString(const char *name, const PortableString & a_string);

   /** Prepends a new int8 to the beginning of a field array in the PortableMessage.
    *  @param name Name of the field to add (or prepend to)
    *  @param val The int8 to add or prepend
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t PrependInt8(const char *name, int8 val);

   /** Prepends a new int16 to the beginning of a field array in the PortableMessage.
    *  @param name Name of the field to add (or prepend to)
    *  @param val The int16 to add or prepend
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t PrependInt16(const char *name, int16 val);

   /** Prepends a new int32 to the beginning of a field array in the PortableMessage.
    *  @param name Name of the field to add (or prepend to)
    *  @param val The int32 to add or prepend
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t PrependInt32(const char *name, int32 val);

   /** Prepends a new int64 to the beginning of a field array in the PortableMessage.
    *  @param name Name of the field to add (or prepend to)
    *  @param val The int64 to add or prepend
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t PrependInt64(const char *name, int64 val);

   /** Prepends a new boolean to the beginning of a field array in the PortableMessage.
    *  @param name Name of the field to add (or prepend to)
    *  @param a_boolean The boolean to add or prepend
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t PrependBool(const char *name, bool a_boolean);

   /** Prepends a new float to the beginning of a field array in the PortableMessage.
    *  @param name Name of the field to add (or prepend to)
    *  @param a_float The float to add or prepend
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t PrependFloat(const char *name, float a_float);

   /** Prepends a new double to the beginning of a field array in the PortableMessage.
    *  @param name Name of the field to add (or prepend to)
    *  @param a_double The double to add or prepend
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t PrependDouble(const char *name, double a_double);

   /** Prepends a new PortableMessage to the beginning of a field array in the PortableMessage..
    *  Note that this method is less efficient than the
    *  AddMessage(PortableMessageRef) method, as this method
    *  necessitates copying all the data in (msg).
    *  @param name Name of the field to add (or prepend to)
    *  @param msg The PortableMessage to add or prepend
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t PrependMessage(const char *name, const PortableMessage & msg);

   /** Prepends a new PortableMessage to the beginning of a field array in the PortableMessage.
    *  Note that this method is more efficient than the
    *  PrependMessage(const PortableMessage &) method, as this method
    *  need not make a copy of the referenced PortableMessage.
    *  @param name Name of the field to add (or prepend to)
    *  @param msgRef Reference to the PortableMessage to add or prepend
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t PrependMessage(const char *name, PortableMessageRef msgRef); 

   /** Prepends a new pointer value to the beginning of a field array in the PortableMessage.
    *  @param name Name of the field to add (or prepend to)
    *  @param ptr The pointer value to add or prepend
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t PrependPointer(const char *name, const void * ptr);

   /** Prepends a new point to the beginning of a field array in the PortableMessage.
    *  @param name Name of the field to add (or prepend to)
    *  @param point The point to add or prepend
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t PrependPoint(const char *name, const PortablePoint & point);

   /** Prepends a new rectangle to the beginning of a field array in the PortableMessage.
    *  @param name Name of the field to add (or prepend to)
    *  @param rect The rectangle to add or prepend
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t PrependRect(const char *name, const PortableRect & rect);

   /** Prepends a new flattened object to the beginning of a field array in the PortableMessage.
    *  @param name Name of the field to add (or prepend to)
    *  @param obj The flattened object to add or prepend
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or a type conflict occurred
    */
   status_t PrependFlat(const char *name, const PortableFlattenable &obj);

   /** Generic method, capable of prepending any type of data to the PortableMessage.
    *  @param name Name of the field to add (or add to)
    *  @param type The B_*_TYPE code indicating the type of data that will be added.
    *  @param data A pointer to the bytes of data to add.  The data is copied out of
    *              this byte buffer, and exactly how it is interpreted depends on (type).
    *              (data) may be NULL, in which case default objects (or uninitialized byte space) is added.
    *  Note:  If you are adding B_MESSAGE_TYPE, then (data) must point to a PortableMessageRef
    *         object, and NOT a PortableMessage object, or flattened-PortableMessage-buffer!
    *         (This is inconsistent with BMessage::AddData, which expects a flattened BMessage buffer.  Sorry!)
    *  @param numBytes The number of bytes that are pointed to by (data).  If (type)
    *                  is a known type such as B_INT32_TYPE or B_RECT_TYPE, (numBytes)
    *                  may be a multiple of the datatype's size, allowing you to add
    *                  multiple entries with a single call.
    *  @return B_NO_ERROR on success, B_ERROR if out of memory or (numBytes) isn't a multiple (type)'s 
    *          known size, or a type conflict occurred.
    */
   status_t PrependData(const char *name, type_code type, const void *data, size_t numBytes);

   /** Removes the (index)'th item from the given field entry.  If the field entry becomes
    *  empty, the field itself is removed also.
    *  @param name Name of the field to remove an item from.
    *  @param index Index of the item to remove.
    *  @return B_NO_ERROR on success, B_ERROR if the field name wasn't found, or the specified index was invalid.
    */
   status_t RemoveData(const char *name, int32 index = 0);

   /** Removes the given field name and its contents from the PortableMessage.
    *  @param name Name of the field to remove.
    *  @return B_NO_ERROR on success, B_ERROR if the field name wasn't found.
    */
   status_t RemoveName(const char *name);

   /** Clears all fields from the PortableMessage. */
   void MakeEmpty();

   /** Retrieve a string value from the PortableMessage.
    *  @param name The field name to look for the string value under.
    *  @param index The index of the string item in its field entry.
    *  @param setMe On success, a pointer to the string value is written into this object.
    *  @return B_NO_ERROR if the string value was found, or B_ERROR if it wasn't.
    */
   status_t FindString(const char *name, int32 index, const char ** setMe) const;

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t FindString(const char *name, const char ** setMe) const {return FindString(name, 0, setMe);}

   /** Retrieve a string value from the PortableMessage.
    *  @param name The field name to look for the string value under.
    *  @param index The index of the string item in its field entry.
    *  @param setMe On success, the value of the string is written into this object.
    *  @return B_NO_ERROR if the string value was found, or B_ERROR if it wasn't.
    */
   status_t FindString(const char *name, int32 index, PortableString & setMe) const;

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t FindString(const char *name, PortableString & setMe) const {return FindString(name, 0, setMe);}

   /** Retrieve an int8 value from the PortableMessage.
    *  @param name The field name to look for the int8 value under.
    *  @param index The index of the int8 item in its field entry.
    *  @param val On success, the value of the int8 is written into this object.
    *  @return B_NO_ERROR if the int8 value was found, or B_ERROR if it wasn't.
    */
   status_t FindInt8(const char *name, int32 index, int8 *val) const;

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t FindInt8(const char *name, int8 *value) const {return FindInt8(name, 0, value);}

   /** Retrieve an int16 value from the PortableMessage.
    *  @param name The field name to look for the int16 value under.
    *  @param index The index of the int16 item in its field entry.
    *  @param val On success, the value of the int16 is written into this object.
    *  @return B_NO_ERROR if the int16 value was found, or B_ERROR if it wasn't.
    */
   status_t FindInt16(const char *name, int32 index, int16 *val) const;

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t FindInt16(const char *name, int16 *value) const {return FindInt16(name, 0, value);}

   /** Retrieve an int32 value from the PortableMessage.
    *  @param name The field name to look for the int32 value under.
    *  @param index The index of the int32 item in its field entry.
    *  @param val On success, the value of the int32 is written into this object.
    *  @return B_NO_ERROR if the int32 value was found, or B_ERROR if it wasn't.
    */
   status_t FindInt32(const char *name, int32 index, int32 *val) const;

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t FindInt32(const char *name, int32 *value) const {return FindInt32(name, 0, value);}

   /** Retrieve an int64 value from the PortableMessage.
    *  @param name The field name to look for the int64 value under.
    *  @param index The index of the int64 item in its field entry.
    *  @param val On success, the value of the int64 is written into this object.
    *  @return B_NO_ERROR if the int64 value was found, or B_ERROR if it wasn't.
    */
   status_t FindInt64(const char *name, int32 index, int64 *val) const;

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t FindInt64(const char *name, int64 *value) const {return FindInt64(name, 0, value);}

   /** Retrieve a boolean value from the PortableMessage.
    *  @param name The field name to look for the boolean value under.
    *  @param index The index of the boolean item in its field entry.
    *  @param value On success, the value of the boolean is written into this object.
    *  @return B_NO_ERROR if the boolean value was found, or B_ERROR if it wasn't.
    */
   status_t FindBool(const char *name, int32 index, bool *value) const;

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t FindBool(const char *name, bool *value) const {return FindBool(name, 0, value);}

   /** Retrieve a float value from the PortableMessage.
    *  @param name The field name to look for the float value under.
    *  @param index The index of the float item in its field entry.
    *  @param f On success, the value of the float is written into this object.
    *  @return B_NO_ERROR if the float value was found, or B_ERROR if it wasn't.
    */
   status_t FindFloat(const char *name, int32 index, float *f) const;

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t FindFloat(const char *name, float *f) const {return FindFloat(name, 0, f);}

   /** Retrieve a double value from the PortableMessage.
    *  @param name The field name to look for the double value under.
    *  @param index The index of the double item in its field entry.
    *  @param d On success, the value of the double is written into this object.
    *  @return B_NO_ERROR if the double value was found, or B_ERROR if it wasn't.
    */
   status_t FindDouble(const char *name, int32 index, double *d) const;

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t FindDouble(const char *name, double *d) const {return FindDouble(name, 0, d);}

   /** Retrieve a PortableMessage value from the PortableMessage.
    *  Note that this method is less efficient than the FindMessage(PortableMessageRef)
    *  method, because it forces the held PortableMessage's data to be copied into (msg)
    *  @param name The field name to look for the PortableMessage value under.
    *  @param index The index of the PortableMessage item in its field entry.
    *  @param msg On success, the value of the PortableMessage is written into this object.
    *  @return B_NO_ERROR if the PortableMessage value was found, or B_ERROR if it wasn't.
    */
   status_t FindMessage(const char *name, int32 index, PortableMessage &msg) const;

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t FindMessage(const char *name, PortableMessage &msg) const {return FindMessage(name, 0, msg);}

   /** Retrieve a PortableMessage value from the PortableMessage.
    *  Note that this method is more efficient than the FindMessage(PortableMessageRef)
    *  method, because it the retrieved PortableMessage need not be copied.
    *  @param name The field name to look for the PortableMessage value under.
    *  @param index The index of the PortableMessage item in its field entry.
    *  @param msgRef On success, the value of the PortableMessage is written into this object.
    *  @return B_NO_ERROR if the PortableMessage value was found, or B_ERROR if it wasn't.
    */
   status_t FindMessage(const char *name, int32 index, PortableMessageRef &msgRef) const;

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t FindMessage(const char *name, PortableMessageRef &msgRef) const {return FindMessage(name, 0, msgRef);}

   /** Retrieve a pointer value from the PortableMessage.
    *  @param name The field name to look for the pointer value under.
    *  @param index The index of the pointer item in its field entry.
    *  @param ptr On success, the value of the pointer is written into this object.
    *  @return B_NO_ERROR if the pointer value was found, or B_ERROR if it wasn't.
    */
   status_t FindPointer(const char *name, int32 index, void ** ptr) const;

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t FindPointer(const char *name, void ** ptr) const {return FindPointer(name, 0, ptr);}

   /** Retrieve a point value from the PortableMessage.
    *  @param name The field name to look for the point value under.
    *  @param index The index of the point item in its field entry.
    *  @param point On success, the value of the point is written into this object.
    *  @return B_NO_ERROR if the point value was found, or B_ERROR if it wasn't.
    */
   status_t FindPoint(const char *name, int32 index, PortablePoint & point) const;

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t FindPoint(const char *name, PortablePoint & point) const {return FindPoint(name, 0, point);}

   /** Retrieve a rectangle value from the PortableMessage.
    *  @param name The field name to look for the rectangle value under.
    *  @param index The index of the rectangle item in its field entry.
    *  @param rect On success, the value of the rectangle is written into this object.
    *  @return B_NO_ERROR if the rectangle value was found, or B_ERROR if it wasn't.
    */
   status_t FindRect(const char *name, int32 index, PortableRect & rect) const;

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t FindRect(const char *name, PortableRect & rect) const {return FindRect(name, 0, rect);}

   /** Retrieve a flattened object from the PortableMessage.
    *  @param name The field name to look for the flattened object value under.
    *  @param index The index of the flattened object item in its field entry.
    *  @param obj On success, the flattened object is copied into this object.
    *  @return B_NO_ERROR if the flattened object was found, or B_ERROR if it wasn't.
    */
   status_t FindFlat(const char *name, int32 index, PortableFlattenable &obj) const;

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t FindFlat(const char *name, PortableFlattenable &obj) const {return FindFlat(name, 0, obj);}

   /** Retrieve a pointer to the raw data bytes of a stored message field of any type.
    *  @param name The field name to retrieve the pointer to
    *  @param type The type_code of the field you are interested, or B_ANY_TYPE if any type is acceptable.
    *  @param data On success, a pointer to the data bytes will be written into this object.
    *  Note:  If you are retrieving B_MESSAGE_TYPE, then (data) will be set to point to a PortableMessageRef
    *         object, and NOT a PortableMessage object, or flattened-PortableMessage-buffer!
    *         (This is inconsistent with BMessage::FindData, which returns a flattened BMessage buffer.  Sorry!)
    *  @param numBytes On success, the number of bytes in the returned data array will be written into this object.
    *  @return B_NO_ERROR if the data pointer was retrieved successfully, or B_ERROR if it wasn't.
    */
   status_t FindData(const char *name, type_code type, int32 index, const void **data, size_t *numBytes) const;

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t FindData(const char *name, type_code type, const void **data, size_t *numBytes) const {return FindData(name, type, 0, data, numBytes);}

   /** As above, only this returns a pointer to an array that you can write directly into, for efficiency.  
    *  You should be very careful with this method, though, as you will be writing into the guts of this PortableMessage.
    *  The returned pointer is not guaranteed to be valid after the PortableMessage is modified by any method!
    *  @see FindData(...)
    */
   status_t FindDataPointer(const char *name, type_code type, int32 index, void **data, size_t *numBytes) const;

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t FindDataPointer(const char *name, type_code type, void **data, size_t *numBytes) const {return FindDataPointer(name, type, 0, data, numBytes);}

   /** Replace a string value in an existing PortableMessage field with a new value.
    *  @param name The field name of an existing field to modify
    *  @param index The index of the entry within the field name to modify
    *  @param newString The new string value to put overwrite the old string with.
    *  @return B_NO_ERROR on success, or B_ERROR if the field wasn't found, or if (index) wasn't a valid index, or out of memory.
    */
   status_t ReplaceString(const char *name, int32 index, const PortableString & newString);

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t ReplaceString(const char *name, const PortableString & newString) {return ReplaceString(name, 0, newString);}

   /** Replace a string value in an existing PortableMessage field with a new value.
    *  @param name The field name of an existing field to modify
    *  @param index The index of the entry within the field name to modify
    *  @param newString The new string value to put overwrite the old string with.
    *  @return B_NO_ERROR on success, or B_ERROR if the field wasn't found, or if (index) wasn't a valid index, or out of memory.
    */
   status_t ReplaceString(const char *name, int32 index, const char * newString) {return ReplaceString(name, index, PortableString(newString));}

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t ReplaceString(const char *name, const char * newString) {return ReplaceString(name, 0, newString);}

   /** Replace an int8 value in an existing PortableMessage field with a new value.
    *  @param name The field name of an existing field to modify
    *  @param index The index of the entry within the field name to modify
    *  @param val The new int8 value to put overwrite the old int8 with.
    *  @return B_NO_ERROR on success, or B_ERROR if the field wasn't found, or if (index) wasn't a valid index, or out of memory.
    */
   status_t ReplaceInt8(const char *name, int32 index, int8 val);

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t ReplaceInt8(const char *name, int8 val) {return ReplaceInt8(name, 0, val);}

   /** Replace an int16 value in an existing PortableMessage field with a new value.
    *  @param name The field name of an existing field to modify
    *  @param index The index of the entry within the field name to modify
    *  @param val The new int16 value to put overwrite the old int16 with.
    *  @return B_NO_ERROR on success, or B_ERROR if the field wasn't found, or if (index) wasn't a valid index, or out of memory.
    */
   status_t ReplaceInt16(const char *name, int32 index, int16 val);

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t ReplaceInt16(const char *name, int16 val) {return ReplaceInt16(name, 0, val);}

   /** Replace an int32 value in an existing PortableMessage field with a new value.
    *  @param name The field name of an existing field to modify
    *  @param index The index of the entry within the field name to modify
    *  @param val The new int32 value to put overwrite the old int32 with.
    *  @return B_NO_ERROR on success, or B_ERROR if the field wasn't found, or if (index) wasn't a valid index, or out of memory.
    */
   status_t ReplaceInt32(const char *name, int32 index, int32 val);

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t ReplaceInt32(const char *name, int32 val) {return ReplaceInt32(name, 0, val);}

   /** Replace an int64 value in an existing PortableMessage field with a new value.
    *  @param name The field name of an existing field to modify
    *  @param index The index of the entry within the field name to modify
    *  @param val The new int64 value to put overwrite the old int8 with.
    *  @return B_NO_ERROR on success, or B_ERROR if the field wasn't found, or if (index) wasn't a valid index, or out of memory.
    */
   status_t ReplaceInt64(const char *name, int32 index, int64 val);

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t ReplaceInt64(const char *name, int64 val) {return ReplaceInt64(name, 0, val);}

   /** Replace a boolean value in an existing PortableMessage field with a new value.
    *  @param name The field name of an existing field to modify
    *  @param index The index of the entry within the field name to modify
    *  @param a_bool The new boolean value to put overwrite the old boolean with.
    *  @return B_NO_ERROR on success, or B_ERROR if the field wasn't found, or if (index) wasn't a valid index, or out of memory.
    */
   status_t ReplaceBool(const char *name, int32 index, bool a_bool);

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t ReplaceBool(const char *name, bool a_bool) {return ReplaceBool(name, 0, a_bool);}

   /** Replace a float value in an existing PortableMessage field with a new value.
    *  @param name The field name of an existing field to modify
    *  @param index The index of the entry within the field name to modify
    *  @param a_float The new float value to put overwrite the old float with.
    *  @return B_NO_ERROR on success, or B_ERROR if the field wasn't found, or if (index) wasn't a valid index, or out of memory.
    */
   status_t ReplaceFloat(const char *name, int32 index, float a_float);

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t ReplaceFloat(const char *name, float a_float) {return ReplaceFloat(name, 0, a_float);}

   /** Replace a double value in an existing PortableMessage field with a new value.
    *  @param name The field name of an existing field to modify
    *  @param index The index of the entry within the field name to modify
    *  @param a_double The new double value to put overwrite the old double with.
    *  @return B_NO_ERROR on success, or B_ERROR if the field wasn't found, or if (index) wasn't a valid index, or out of memory.
    */
   status_t ReplaceDouble(const char *name, int32 index, double a_double);

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t ReplaceDouble(const char *name, double a_double) {return ReplaceDouble(name, 0, a_double);}

   /** Replace a pointer value in an existing PortableMessage field with a new value.
    *  @param name The field name of an existing field to modify
    *  @param index The index of the entry within the field name to modify
    *  @param ptr The new pointer value to put overwrite the old pointer with.
    *  @return B_NO_ERROR on success, or B_ERROR if the field wasn't found, or if (index) wasn't a valid index, or out of memory.
    */
   status_t ReplacePointer(const char *name, int32 index, const void * ptr);

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t ReplacePointer(const char *name, const void * ptr) {return ReplacePointer(name, 0, ptr);}

   /** Replace a point value in an existing PortableMessage field with a new value.
    *  @param name The field name of an existing field to modify
    *  @param index The index of the entry within the field name to modify
    *  @param point The new point value to put overwrite the old point with.
    *  @return B_NO_ERROR on success, or B_ERROR if the field wasn't found, or if (index) wasn't a valid index, or out of memory.
    */
   status_t ReplacePoint(const char *name, int32 index, const PortablePoint & point);

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t ReplacePoint(const char *name, const PortablePoint & point) {return ReplacePoint(name, 0, point);}

   /** Replace a rectangle value in an existing PortableMessage field with a new value.
    *  @param name The field name of an existing field to modify
    *  @param index The index of the entry within the field name to modify
    *  @param rect The new rectangle value to put overwrite the old rectangle with.
    *  @return B_NO_ERROR on success, or B_ERROR if the field wasn't found, or if (index) wasn't a valid index, or out of memory.
    */
   status_t ReplaceRect(const char *name, int32 index, const PortableRect & rect);

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t ReplaceRect(const char *name, const PortableRect & rect) {return ReplaceRect(name, 0, rect);}

   /** Replace a PortableMessage value in an existing PortableMessage field with a new value.
    *  @param name The field name of an existing field to modify
    *  @param index The index of the entry within the field name to modify
    *  @param msg The new PortableMessage value to put overwrite the old PortableMessage with.
    *  @return B_NO_ERROR on success, or B_ERROR if the field wasn't found, or if (index) wasn't a valid index, or out of memory.
    */
   status_t ReplaceMessage(const char *name, int32 index, const PortableMessage &msg);

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t ReplaceMessage(const char *name, const PortableMessage &msg) {return ReplaceMessage(name, 0, msg);}

   /** Replace a PortableMessage value in an existing PortableMessage field with a new value.
    *  @param name The field name of an existing field to modify
    *  @param index The index of the entry within the field name to modify
    *  @param msgRef The new PortableMessage value to put overwrite the old PortableMessage with.
    *  @return B_NO_ERROR on success, or B_ERROR if the field wasn't found, or if (index) wasn't a valid index, or out of memory.
    */
   status_t ReplaceMessage(const char *name, int32 index, PortableMessageRef msgRef);

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t ReplaceMessage(const char *name, PortableMessageRef msgRef) {return ReplaceMessage(name, 0, msgRef);}

   /** Replace a flattened object in an existing PortableMessage field with a new object.
    *  @param name The field name of an existing field to modify
    *  @param index The index of the entry within the field name to modify
    *  @param obj The new flattened object value to put overwrite the old flattened object with.
    *  @return B_NO_ERROR on success, or B_ERROR if the field wasn't found, or if (index) wasn't a valid index, or out of memory.
    */
   status_t ReplaceFlat(const char *name, int32 index, const PortableFlattenable &obj);

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t ReplaceFlat(const char *name, const PortableFlattenable &obj) {return ReplaceFlat(name, 0, obj);}

   /** Replace one entry in a field of any type.
    *  @param name The field name to replace an entry in.
    *  @param type The type_code of the field you are interested, or B_ANY_TYPE if any type is acceptable.
    *  @param data The bytes representing the item you wish to replace an old item in the field with
    *  Note:  If you are replacing B_MESSAGE_TYPE, then (data) must point to a PortableMessageRef
    *         object, and NOT a PortableMessage object, or flattened-PortableMessage-buffer!
    *         (This is inconsistent with BMessage::ReplaceData, which expects a flattened BMessage buffer.  Sorry!)
    *  @param numBytes The number of bytes in your data array.
    *  @return B_NO_ERROR if the data item(s) were replaced successfully, or B_ERROR they weren't.
    */
   status_t ReplaceData(const char *name, type_code type, int32 index, const void *data, size_t data_size);

   /** As above, only (index) isn't specified.  It is assumed to be zero. */
   status_t ReplaceData(const char *name, type_code type, const void *data, size_t data_size) {return ReplaceData(name, type, 0, data, data_size);}

   /** Returns true iff there is a field with the given name and type present in the PortableMessage
    *  @param fieldName the field name to look for.
    *  @param type the type to look for, or B_ANY_TYPE if type isn't important to you.
    *  @return true if such a field exists, else false. 
    */
   bool HasName(const char * fieldName, type_code type = B_ANY_TYPE) const;

   /** Take the data under (name) in this message, and moves it into (moveTo). 
    *  Any data that was under (name) in (moveTo) will be replaced.
    *  @param name Name of an existing field to be moved.
    *  @param moveTo A PortableMessage to move the field into.
    *  @result B_NO_ERROR on success, or B_ERROR if there is an error moving the field.
    */
   status_t MoveName(const char * name, PortableMessage &moveTo);

   /**
    * Take the data under (name) in this message, and copies it into (moveTo). 
    * Any data that was under (name) in (moveTo) will be replaced.
    * @param name Name of an existing field to be copied.
    * @param copyTo A PortableMessage to copy the field into.
    *  @result B_NO_ERROR on success, or B_ERROR if there is an error copying the field.
    */
   status_t CopyName(const char * name, PortableMessage &copyTo) const;

   /**
    * Returns an iterator that iterates over the names of the fields in this
    * message.  If (type) is B_ANY_TYPE, then all field names will be included
    * in the traversal.  Otherwise, only names of the given type will be included.
    * @param type Type of fields you wish to iterate over, or B_ANY_TYPE to iterate over all fields.
    * @return A PortableMessageFieldNameIterator object.
    */
   PortableMessageFieldNameIterator GetFieldNameIterator(type_code type = B_ANY_TYPE) const;

   /**
    * This method sets the pools that the PortableMessage class uses by default in situations where
    * now pools are available (e.g. from a passed-in PortableMessageRef or somesuch).  By default,
    * no pooling is done in these situations.  For a minor CPU-efficiency boost in single-threaded
    * programs, you can call SetInternalDefaultPools(GetMessagePool(), GetMessageRefPool()).  Don't
    * do this for multi-threaded programs, though, as it will cause race conditions!  
    * (For multi-threaded Be programs, you can use MessageTransceiverThread::GetMessagePool() and
    * MessageTransceiverThread::GetMessageRefPool() instead, as these will do appropriate locking
    * to guarantee thread-safety)
    * @param messagePool an ObjectPool to recycle PortableMessages into
    * @param messageRefPool an ObjectPool to recycle PortableMessageRef::RefPools into
    */
   static void SetInternalDefaultPools(PortableMessageRef::ItemPool * messagePool, PortableMessageRef::RefPool *messageRefPool);

private:
   // Helper functions!

   status_t AddData(const char *name, type_code type, const void *data, size_t numBytes, bool prepend);

   // Given a known type_code, returns the size of an item of that type.
   // Returns zero if items of the given type are variable length.
   status_t GetElementSize(type_code type, size_t & setSize) const;

   AbstractDataArray * GetArray(const char * arrayName, type_code etc) const {return GetArray(PortableString(arrayName), etc);}
   AbstractDataArray * GetArray(const PortableString & arrayName, type_code etc) const;

   AbstractDataArray * GetOrCreateArray(const char * arrayName, type_code tc) {return GetOrCreateArray(PortableString(arrayName), tc);}
   AbstractDataArray * GetOrCreateArray(const PortableString & arrayName, type_code tc);

   status_t AddFlatAux(const char * name, const PortableFlattenable & flat, type_code etc, bool prepend);
   status_t AddDataAux(const char * name, const uint8 * data, size_t size, type_code etc, bool prepend);

   status_t FindFlatAux(const char * name, int32 index, PortableFlattenable & flat) const;
   status_t FindDataAux(const char * name, int32 index, const uint8 ** setData, size_t * setSize, type_code etc) const;
   status_t FindDataItemAux(const char * name, int32 index, type_code tc, void * setValue, size_t valueSize) const;

   status_t ReplaceFlatItemAux(const char * name, int32 index, const PortableFlattenable & flat, type_code tc);
   status_t ReplaceDataAux(const char * name, int32 index, uint8 * dataBuf, size_t bufSize, type_code tc);
 
   // Iterator support methods
   friend class PortableMessageFieldNameIterator;
   const char * GetNextFieldName(_eitr * iter, type_code tc) const;
   const char * PeekNextFieldName(_eitr * iter, type_code tc) const;
   PortableHashtable<PortableString, AbstractDataArray *> _entries;   
};

#endif /* _PORTABLEMESSAGE_H */

