/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */  

#include <new>
#include <typeinfo>
#include <string.h>

#ifndef __BEOS__
#include <arpa/inet.h>
#endif

#include <netdb.h>

#include "reflector/ReflectServer.h"
#include "netutil/NetworkUtilityFunctions.h"

static uint32 _outOfMemoryWarningCount = 0L;
static uint32 _maxAllocableBytes = (uint32)-1;

static uint32 _bytesAlloced = 0L;

static PortableQueue<OutOfMemoryHandlerFunc> _outOfMemoryHandlers;

void AddOutOfMemoryHandler(OutOfMemoryHandlerFunc f)
{
   _outOfMemoryHandlers.AddTail(f);
}

uint32 GetNumAvailableBytes()
{
   return (_bytesAlloced < _maxAllocableBytes) ? _maxAllocableBytes-_bytesAlloced : 0;
}
                                                              
uint32 GetNumUsedBytes()
{
   return _bytesAlloced;
}

uint32 GetMaxNumBytes()
{
   return _maxAllocableBytes;
}

// Metrowerk's compiler crashes at run-time if this memory-tracking
// code is enabled.  I haven't been able to figure out why (everything
// looks okay, but after a few dozen malloc()/free() calls, free()
// crashes!) so I'm just going to disable this code for PowerPC/Metrowerks
// machines.  Sorry!   --jaf 12/01/00
#ifdef __MWERKS__
#define DISABLE_MUSCLE_MEMORY_TRACKING
#endif

#ifndef DISABLE_MUSCLE_MEMORY_TRACKING

static void * jallocAux(size_t);
void * jallocAux(size_t s)
{
    if (_bytesAlloced + s > _maxAllocableBytes) return NULL;

    uint32 allocSize = s + sizeof(uint32);  // requested size, plus 4 bytes for our use
    uint32 * a = (uint32*) malloc(allocSize);
    if (a)
    {
       *a = allocSize;  // a tag so jfree() will know how big the alloc was
       _bytesAlloced += allocSize;
       return a+1;  // user doesn't see it, of course
    }
    else return NULL;
}

static void * jalloc(size_t);
void * jalloc(size_t s)
{
   void * ret = jallocAux(s);
   if (ret == NULL)
   {
      GetMessagePool()->Drain();     // dump some unnecessary objects to
      GetMessageRefPool()->Drain();  // free up memory
      for (int i=_outOfMemoryHandlers.CountItems()-1; i>=0; i--) _outOfMemoryHandlers[i]();  // user-defined lo-mem callbacks
      ret = jallocAux(s);
   }
   if (ret == NULL) _outOfMemoryWarningCount++;  // still no mem?  Okay, time to break out the whoop ass
   return ret;
}

static void jfree(void *);
void jfree(void * p)
{
    uint32 * realStart = ((uint32*)p)-1;
    uint32 s = *realStart;
    _bytesAlloced -= s;
    free(realStart);
}

void* operator new(size_t s) throw (bad_alloc)
{
   void * ret = jalloc(s);
   if (ret == NULL) throw bad_alloc();
   return ret;
}

void* operator new[](size_t s) throw (bad_alloc)
{
   void * ret = jalloc(s);
   if (ret == NULL) throw bad_alloc();
   return ret;
}

void* operator new(size_t s, nothrow_t const &) throw ()
{
   return jalloc(s);
}

void* operator new[](size_t s, nothrow_t const &) throw ()
{
   return jalloc(s);
}

void operator delete(void* p) throw()
{
   if (p) jfree(p);
}

void operator delete[](void* p) throw()
{
   if (p) jfree(p);
}

#endif

status_t
ReflectServer ::
AddNewSession(AbstractReflectSessionRef ref, int socket)
{
   AbstractReflectSession * newSession = ref.GetItemPointer();
   if (newSession == NULL) return B_ERROR;

   newSession->_socket = socket;
   newSession->_owner = this;  // in case CreateGateway() needs to use the owner

   if (socket >= 0)
   {
      newSession->_gateway = newSession->CreateGateway(socket);
      if (newSession->_gateway == NULL) 
      {
         newSession->_socket = -1;
         return B_ERROR;
      }

      uint32 remoteIP;
      uint16 remotePort;
      if (GetPeerInfo(socket, remoteIP, remotePort) == B_NO_ERROR)
      {
         newSession->_port = remotePort;

         char ipBuf[13];
         Inet_NtoA(remoteIP, ipBuf);
         if (strcmp(ipBuf, "127.0.0.1") == 0)
         {
            // 127.0.0.1 isn't good enough (other clients can't use it to connect to)... so find out our actual name
            char buf[512];
            (void) gethostname(buf, sizeof(buf));
            newSession->_hostName = buf;
         }
         else newSession->_hostName = ipBuf;
      }
      else 
      {
         newSession->_hostName = "<unknown>";
         newSession->_port     = 0;
      }
   }
   else
   {
      newSession->_hostName = "<local>";
      newSession->_port     = 0;
   }

   return AddNewSession(ref);
}


status_t
ReflectServer ::
AddNewSession(AbstractReflectSessionRef ref)
{
   AbstractReflectSession * newSession = ref.GetItemPointer();
   if (newSession == NULL) return B_ERROR;
   newSession->_owner = this;

   if (_sessions.AddTail(ref))
   {
      _idLookup.Put(newSession->GetSessionIDString(), ref);
      if (newSession->AttachedToServer() == B_NO_ERROR)
      {
         LogTime(MUSCLE_LOG_INFO, "New %s [%s] from [%s:%i] (%lu total)\n", newSession->GetTypeName(), newSession->GetSessionIDString(), newSession->GetHostName(), newSession->GetClientPort(), _sessions.CountItems());
         return B_NO_ERROR;
      }
      else 
      {
         LogTime(MUSCLE_LOG_INFO, "%s [%s] from [%s:%i] aborted startup (%lu total)\n", newSession->GetTypeName(), newSession->GetSessionIDString(), newSession->GetHostName(), newSession->GetClientPort(), _sessions.CountItems());
         _sessions.RemoveTail();
         _idLookup.Remove(newSession->GetSessionIDString());
      }
   }

   return B_ERROR;
}


ReflectServer :: ReflectServer(uint32 maxBytes) : _idLookup(CStringHashFunc), _keepServerGoing(true)
{
   _maxAllocableBytes = maxBytes;
   _nextPulseAt = MUSCLE_NEVER_PULSE;

   // This call lets the PortableMessage class use its default message pools for its
   // internal purposes.  It's safe to allow this here, since all PortableMessage calls
   // will be made from the same thread.  For a multi-threaded system, don't do this...
   PortableMessage::SetInternalDefaultPools(GetMessagePool(), GetMessageRefPool());

   _idLookup.SetCompareFunction(CStringCompareFunc);
}

ReflectServer :: ~ReflectServer()
{
   // empty
}

void
ReflectServer :: Cleanup()
{
   _idLookup.Clear();

   // This silly little contortion is necessary to keep the
   // StorageReflectSession destructors from accessing themselves 
   // in the _sessions list while the _sessions list is being cleared!
   // (i.e. we keep an additional reference of each session around in
   // the (temp) array, so that we can clear _sessions without activating
   // all the destructors.  Then when we clear (temp), the destructors are
   // activated, but the _sessions list is already safely cleared)
   PortableQueue<AbstractReflectSessionRef> temp;
   for (int i=_sessions.CountItems()-1; i>=0; i--) temp.AddTail(_sessions[i]);

   _sessions.MakeEmpty();
   _lameDuckSessions.MakeEmpty();
   temp.MakeEmpty();
}

status_t
ReflectServer ::
ReadyToRun()
{
   return B_NO_ERROR;
}

const char *
ReflectServer ::
GetServerName() const
{
   return "MUSCLE";
}

status_t 
ReflectServer ::
ServerProcessLoop(ReflectSessionFactoryFunc createSessionFunc, short acceptPort)
{
   // Setup the connection-accepting socket
   struct sockaddr_in saSocket;
   const long trueValue = 1;
   memset(&saSocket, 0, sizeof(saSocket));
   saSocket.sin_family      = AF_INET;
   saSocket.sin_addr.s_addr = htonl(INADDR_ANY);
   saSocket.sin_port        = htons(acceptPort);

   int acceptSocket;
   if ((acceptSocket = socket(AF_INET, SOCK_STREAM, 0)) < 0)
   {
      LogTime(MUSCLE_LOG_CRITICALERROR, "Error creating accept socket!\n");
      return B_ERROR;
   }
   (void) setsockopt(acceptSocket, SOL_SOCKET, SO_REUSEADDR, &trueValue, sizeof(long));
   if ((bind(acceptSocket, (struct sockaddr *) &saSocket, sizeof(saSocket)))||(listen(acceptSocket, 20)))
   {
      LogTime(MUSCLE_LOG_CRITICALERROR, "Error binding/listening on accept socket!\n");
      closesocket(acceptSocket);
      return B_ERROR;
   }

   LogTime(MUSCLE_LOG_DEBUG, "This %s server was compiled on " __DATE__ " " __TIME__ "\n", GetServerName());
   LogTime(MUSCLE_LOG_DEBUG, "The server was compiled against MUSCLE version %s\n", MUSCLE_VERSION_STRING);
   if (ReadyToRun() != B_NO_ERROR) 
   {
      LogTime(MUSCLE_LOG_CRITICALERROR, "Server:  ReadyToRun() failed, aborting.\n");
      return B_ERROR;
   }

   LogTime(MUSCLE_LOG_INFO, "%s Server is active and listening on port %i\n",GetServerName(), acceptPort);
   fd_set readSet;
   fd_set writeSet;
   while(_keepServerGoing)
   {
      // Remove any sessions that were marked for removal
      // (Done first since this may have occurred even in ReadyToRun())
      {
         AbstractReflectSessionRef duckRef;
         while(_lameDuckSessions.RemoveHead(duckRef))
         {
            AbstractReflectSession * duck = duckRef.GetItemPointer();
            if (duck)
            {
               for (int j=_sessions.CountItems()-1; j>=0; j--)
               {
                  AbstractReflectSession * next = _sessions.GetItemPointer(j)->GetItemPointer();
                  if (next == duck)
                  {
                     const char * id = next->GetSessionIDString();
                     LogTime(MUSCLE_LOG_INFO, "Closed %s [%s] to [%s:%i] (%lu total)\n", next->GetTypeName(), id, next->GetHostName(), next->GetClientPort(), _sessions.CountItems()-1);
                     _idLookup.Remove(id);
                     _sessions.RemoveItemAt(j);
                     break;
                  }
               }
            }
         }
      }

      FD_ZERO(&readSet);
      FD_ZERO(&writeSet);
      FD_SET(acceptSocket, &readSet);

      // Find out which of our sessions wants to write data
      int maxSocket = acceptSocket;
      {
         for (int i=_sessions.CountItems()-1; i>=0; i--)
         {
            AbstractReflectSession * session = _sessions.GetItemPointer(i)->GetItemPointer();
            if ((session)&&(session->_gateway))
            {
               int sessionSocket = session->_socket;
               if (sessionSocket >= 0)
               {
                  bool in  = session->_gateway->GetInputEnabled(); 
                  bool out = session->_gateway->HasBytesToOutput();     
                  if (in)  FD_SET(sessionSocket, &readSet);
                  if (out) FD_SET(sessionSocket, &writeSet);
                  if (((in)||(out))&&(sessionSocket > maxSocket)) maxSocket = sessionSocket;
               }
            }
         }
      }

      // Calculate timeout for Pulse(), if any
      struct timeval waitTime;
      {
         const uint64 now64 = GetCurrentTime();
         _nextPulseAt = GetPulseTime(now64);

         // If we are now expecting a Pulse(), calculate how long to wait for it.
         if (_nextPulseAt != MUSCLE_NEVER_PULSE)
         {
            // Calculate how long to wait until the pulse time
            uint64 wt = (_nextPulseAt > now64) ? (_nextPulseAt - now64) : 0;
            Convert64ToTimeVal(wt, waitTime);
	    MASSERT((waitTime.tv_sec >= 0), "Negative select() timeout!");
         }
      }
 
      int numBitsSet = select(maxSocket+1, &readSet, &writeSet, NULL, (_nextPulseAt == MUSCLE_NEVER_PULSE) ? NULL : &waitTime);
      if (numBitsSet < 0) 
      {
         LogTime(MUSCLE_LOG_CRITICALERROR, "select() failed, aborting!\n");
         closesocket(acceptSocket);
         return B_ERROR;
      }

      if (_nextPulseAt != MUSCLE_NEVER_PULSE)
      {
         const uint64 now64 = GetCurrentTime();
         if (now64 >= _nextPulseAt)
         {
            Pulse(now64);
            _nextPulseAt = MUSCLE_NEVER_PULSE;
         }
      }

      if (FD_ISSET(acceptSocket, &readSet)) 
      {
         // Accept new connection and start up a session for it
#if __BEOS__ || __APPLE__
         int nLen = sizeof(saSocket);
#else
         size_t nLen = sizeof(saSocket);
#endif
         int newSocket = accept(acceptSocket, (struct sockaddr *)&saSocket, &nLen);
         if (newSocket >= 0)
         {
            AbstractReflectSession * newSession = createSessionFunc();
            if (newSession) 
            {
               AbstractReflectSessionRef newSessionRef(newSession, NULL, NULL);
               if (AddNewSession(newSessionRef, newSocket) != B_NO_ERROR) closesocket(newSocket);
            }
            else
            {
               LogTime(MUSCLE_LOG_ERROR, "Error, couldn't create new reflector session!\n");
               closesocket(newSocket);
            }
         }
         else LogTime(MUSCLE_LOG_WARNING, "Warning, accept() failed!\n");
      }

      // Do the session I/Os...
      {
         for (int i=_sessions.CountItems()-1; i>=0; i--)
         {
            AbstractReflectSession * session = _sessions.GetItemPointer(i)->GetItemPointer();
            if ((session)&&(session->_gateway))
            {
               AbstractMessageIOGateway & gateway = *(session->_gateway);
               int socket = session->_socket;
               if (socket >= 0)
               {
                  bool readError  = (FD_ISSET(socket, &readSet))&&(gateway.DoInput()   != B_NO_ERROR);
                  bool writeError = (FD_ISSET(socket, &writeSet))&&(gateway.DoOutput() != B_NO_ERROR); 
                  if ((readError)||(writeError)) 
                  {
                     if (session->ClientConnectionClosed()) _lameDuckSessions.AddTail(_sessions[i]);
                     else
                     {
                        LogTime(MUSCLE_LOG_INFO, "TCP Connection to [%s:%i] was severed.\n", session->GetHostName(), session->GetClientPort());
                        session->_socket = -1;  // so we won't try to use it
                     }
                  }
               }
            }
         }
      }

      // Let the various sessions handle their messages...
      while(true)
      {
         bool loopAgain = false;

         PortableMessageRef nextMessageRef;
         int numSessions = _sessions.CountItems();
         for (int i=0; ((loopAgain == false)&&(i<numSessions)); i++)
         {
            AbstractReflectSessionRef sRef = _sessions[i];  // used to keep session from deleting itself prematurely
            AbstractReflectSession * s = sRef.GetItemPointer();

            if (s)
            {
               // Note:  It *is* necessary to check fromSession->_gateway every time through this
               //        loop, because a ReplaceSession() call might set it to NULL.
               _outOfMemoryWarningCount = 0;  // (s)'s responsibility for starts here!  If we run out of mem, he's history
               while((s->_gateway)&&(s->_gateway->GetNextIncomingMessage(nextMessageRef) == B_NO_ERROR)) 
               {
                  s->MessageReceivedFromGateway(nextMessageRef);
                  if (_outOfMemoryWarningCount > 0) break;  // big trouble!
                  if (s->_gateway == NULL)
                  {
                     // oops!  This session must have replaced itself.  Start through the loop again.
                     loopAgain = true;
                     break;
                  }
               }
               if (_outOfMemoryWarningCount > 0)
               {
                  LogTime(MUSCLE_LOG_CRITICALERROR, "Low Memory!  Aborting session [%s:%i] to get some back!\n", s->GetHostName(), s->GetClientPort());
                  _lameDuckSessions.AddTail(sRef);  // of course, if there is no mem to add this, we're REALLY screwed!  
                  _outOfMemoryWarningCount = 0;
                  break;
               }
            }
         }

         if (loopAgain == false) break;
      }
   }
   return B_NO_ERROR;
}

void 
ReflectServer ::
Pulse(uint64 now)
{
   for (int i=_sessions.CountItems()-1; i>=0; i--)
   {
      AbstractReflectSession * session = _sessions.GetItemPointer(i)->GetItemPointer();
      if ((session)&&(session->_gateway))
      {
         AbstractMessageIOGateway * g = session->_gateway;
         if (now >= g->_nextPulseAt)
         {
            g->Pulse(now);
            g->_nextPulseAt = MUSCLE_NEVER_PULSE;
         }
      }
   }
}

uint64
ReflectServer ::
GetPulseTime(uint64 now)
{
   uint64 rt = MUSCLE_NEVER_PULSE;

   for (int i=_sessions.CountItems()-1; i>=0; i--)
   {
      AbstractReflectSession * session = _sessions.GetItemPointer(i)->GetItemPointer();
      if ((session)&&(session->_gateway))
      {
         AbstractMessageIOGateway * g = session->_gateway;
         if (g->_nextPulseAt == MUSCLE_NEVER_PULSE) g->_nextPulseAt = g->GetPulseTime(now);
         if (g->_nextPulseAt < rt) rt = g->_nextPulseAt;
      }
   }
   return rt;
}


PortableQueue<AbstractReflectSessionRef> & 
ReflectServer ::
GetSessionsList()
{
   return _sessions;
}


status_t
ReflectServer ::
ReplaceSession(AbstractReflectSessionRef newSessionRef, AbstractReflectSession * oldSession)
{
   // move the gateway from the old session to the new one...
   AbstractReflectSession * newSession = newSessionRef.GetItemPointer();
   if (newSession == NULL) return B_ERROR;

   newSession->_gateway  = oldSession->_gateway;
   newSession->_hostName = oldSession->_hostName;
   newSession->_port     = oldSession->_port;
   newSession->_socket   = oldSession->_socket;

   if (AddNewSession(newSessionRef) == B_NO_ERROR)
   {
      oldSession->_gateway = NULL;   /* gateway now belongs to newSession */
      EndSession(oldSession);
      return B_NO_ERROR;
   }
   else
   {
       // Oops, rollback changes and error out
       newSession->_gateway  = NULL;
       newSession->_hostName = "";
       newSession->_port     = 0; 
       newSession->_socket   = -1;
       return B_ERROR;
   }
}


void
ReflectServer ::
EndSession(AbstractReflectSession * who)
{
   for (int i=_sessions.CountItems()-1; i>=0; i--)
   {
      if (_sessions.GetItemPointer(i)->GetItemPointer() == who)
      {
         _lameDuckSessions.AddTail(_sessions[i]);
         return;
      }
   }
}

void
ReflectServer ::
EndServer()
{
   _keepServerGoing = false;
}
