/* This file is Copyright 2001 Level Control Systems.  See the included LICENSE.txt file for details. */

#ifndef PORTABLE_REFLECT_SERVER_H
#define PORTABLE_REFLECT_SERVER_H

#include "reflector/AbstractReflectSession.h"

/** Returns the number of bytes available for the MUSCLE server to allocate from the heap */
uint32 GetNumAvailableBytes();

/** Returns the number of bytes the MUSCLE server currently has allocated */
uint32 GetNumUsedBytes();

/** Returns the maximum number of bytes the MUSCLE server will allow itself to allocate. */
uint32 GetMaxNumBytes();

typedef void (*OutOfMemoryHandlerFunc)();

/** Add a callback function that will be called when we are running out of memory.
 *  Functions added to this list can try to free memory, place an order at Chip Merchant, or whatever.
 */
extern void AddOutOfMemoryHandler(OutOfMemoryHandlerFunc f);

/** This class represents a MUSCLE server:  It runs on a centrally located machine,
 *  and many clients may connect to it simultaneously.  This server can then redirect messages
 *  uploaded by any client to other clients in a somewhat efficient manner.
 *  This class can be used as-is, or subclassed if necessary.
 *  There is typically only one ReflectServer object present in a given MUSCLE server program.
 */
class ReflectServer
{
public: 
   /** Constructor.
    *  @param maxBytes Specifies the maximum number of bytes of memory the server will allow itself 
    *                  to have allocated at once.  Defaults to its maximum value of 4 gigabytes.
    */
   ReflectServer(uint32 maxBytes = ((uint32)-1L));

   /** Destructor.  */
   virtual ~ReflectServer();

   /** The main loop for the message reflection server.
    *  This method will not return until the server stops running (usually due to an error).
    *  @param createSessionFunc should be a function of the form:   AbstractReflectSession * MyFunc();
    *                           that returns a new AbstractReflectSession object to be used for communicating with 
    *                           a given socket.  The object this function returns largely determines the behaviour 
    *                           of the server.  Most non-abstract ReflectSession classes should come with such a 
    *                           function included.
    *  @param acceptPort The TCP port the server will listen on.  (Traditionally 2960)
    *  @return B_NO_ERROR if the server has decided to exit peacefully, or B_ERROR if there was a 
    *                     fatal error during setup or execution.
    */
   virtual status_t ServerProcessLoop(ReflectSessionFactoryFunc createSessionFunc, short acceptPort = 2960);

   /**
    * Called after the server is set up, but just before accepting any connections.
    * Should return B_NO_ERROR if it's okay to continue, or B_ERROR to abort and shut down the server.
    * @return Default implementation returns B_NO_ERROR.
    */
   virtual status_t ReadyToRun();

   /**
    * Should be called just before the ReflectServer is to be destroyed.
    * this in a good place to put any cleanup code.  Be sure
    * to call Cleanup() of your parent class as well!
    * (We can't just do this in the destructor, as some cleanup
    * relies on the subclass still being functional, which it isn't
    * when our destructor gets called!)
    */
   virtual void Cleanup();

   /** Accessor to our central-state repository message */
   PortableMessage & GetCentralState() {return _centralState;}

protected:
   /**
    * Adds a new session that uses the given socket for I/O.
    * @param ref New session to add to the server.
    * @param socket The TCP socket that the new session will be useing, or -1, if the new session has no TCP connection.
    * @return B_NO_ERROR if the new session was added successfully, or B_ERROR if there was an error setting it up.
    */
   virtual status_t AddNewSession(AbstractReflectSessionRef ref, int socket);

   /**
    * This version of AddNewSession (which is called by the previous 
    * version) assumes that the gateway, hostname, port, etc of the
    * new session have already been set up.
    * @param ref The new session to add.
    * @return B_NO_ERROR if the new session was added successfully, or B_ERROR if there was an error setting it up.
    */
   virtual status_t AddNewSession(AbstractReflectSessionRef ref);

   /**
    * Returns a human-readable string that describes the type of server that is running.  
    * @return Default implementation returns "MUSCLE".
    */
   virtual const char * GetServerName() const;

   /**
    * May be overridden to specify when Pulse() should next be called.
    * @param now is the current time, in microseconds, for reference.
    * @return Should return the time at which you wish to have Pulse() called next.
    *         Defaults to returning PR_NEVER_PULSE, which indicates that no Pulse()
    *         call should be scheduled.
    */
   virtual uint64 GetPulseTime(uint64 now);

   /**
    * If GetPulseTime() was overridden to return a value other than MUSCLE_NEVER_PULSE, 
    * then this callback will be called at (or shortly after) the time that GetPulseTime() returned.
    * @param now is the current time of day, for reference.  
    * Default implementation calls Pulse() on any i/o gateways that scheduled Pulse() events.
    */
   virtual void Pulse(uint64 now);

   /** Returns our contained list of active sessions.  */
   PortableQueue<AbstractReflectSessionRef> & GetSessionsList();

   /** Called by a session to get itself removed & destroyed */
   void EndSession(AbstractReflectSession * which);

   /** Call this and the server will quit ASAP */
   void EndServer();

   /**
    * Called by a session to get itself replaced (the
    * new session will continue using the same message io streams
    * as the old one)
    * @return B_NO_ERROR on success, B_ERROR if the new session
    * returns an error in its AttachedToServer() method.  If 
    * B_ERROR is returned, then this call is guaranteed not to
    * have had any effect on the old session.
    */
   status_t ReplaceSession(AbstractReflectSessionRef newSession, AbstractReflectSession * replaceThisOne);

private:
   friend class AbstractReflectSession;

   PortableMessage _centralState;
   PortableQueue<AbstractReflectSessionRef> _sessions;
   PortableHashtable<const char *, AbstractReflectSessionRef> _idLookup;
   PortableQueue<AbstractReflectSessionRef> _lameDuckSessions;
   uint64 _nextPulseAt;
   bool _keepServerGoing;
};

#endif
