/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */  

#include <stdio.h>
#include <stdlib.h>
#include "reflector/ReflectServer.h"
#include "reflector/DumbReflectSession.h"
#include "reflector/StorageReflectSession.h"
#include "stringtokenizer/StringTokenizer.h"
#include "regex/StringMatcher.h"

#define DEFAULT_MUSCLED_PORT 2960

int main(int argc, char ** argv) 
{
   int port = (argc > 1) ? atoi(argv[1]) : 0;
   uint32 maxBytes = (uint32)-1L;
   PortableQueue<PortableString> banList;
   PortableMessage tempPrivs;

   // Parse arguments
   for (int i=1; i<argc; i++)
   {
      StringTokenizer tok(argv[i], (const char *)"=");
      const char * keyword = tok.GetNextToken();

      if (keyword == NULL) keyword = argv[i];
      if (keyword)
      {
         const char * value = tok.GetNextToken();
         if (value == NULL) value = "";

         while(keyword[0] == '-') keyword++;  // stupid Unix geeks anyway
         if (strcmp(keyword, "help") == 0)
         {
            Log(MUSCLE_LOG_INFO, "Usage:  muscled [daemon] [port=2960] [maxmem=512] [display=lvl] [log=lvl]\n");
            Log(MUSCLE_LOG_INFO, " - If daemon is specified, muscled will run as a background process\n");
            Log(MUSCLE_LOG_INFO, " - port may be any number between 1 and 65536\n");
            Log(MUSCLE_LOG_INFO, " - maxmem is the max megabytes of memory the server may use\n");
            Log(MUSCLE_LOG_INFO, " - lvl is: none, critical, errors, warnings, info, debug, or trace.\n");
            Log(MUSCLE_LOG_INFO, " - You may also put one or more ban=<pattern> arguments in.\n");
            Log(MUSCLE_LOG_INFO, "   Each pattern specifies one or more IP addresses to\n");
            Log(MUSCLE_LOG_INFO, "   disallow connections from, e.g. ban=192.168.*.*\n");
            Log(MUSCLE_LOG_INFO, " - To assign privileges, specify one of the following:\n");
            Log(MUSCLE_LOG_INFO, "   privban=<pattern>, privunban=<pattern>, privkick=<pattern>\n");
            Log(MUSCLE_LOG_INFO, "   or privall=<pattern>.  privall assigns all privileges to the\n");
            Log(MUSCLE_LOG_INFO, "   matching IP addresses.\n");
            exit(5);
         }
         else if (value)
         {
                 if (strcmp(keyword, "port") == 0) port = atoi(value);
            else if (strcmp(keyword, "daemon") == 0)
            {
               LogTime(MUSCLE_LOG_INFO, "Spawning off a muscled daemon-child...\n");
               if (BecomeDaemonProcess() != B_NO_ERROR)
               {
                  LogTime(MUSCLE_LOG_CRITICALERROR, "Couldn't spawn!\n");
                  exit(10);
               }
            }
            else if (strcmp(keyword, "maxmem") == 0) 
            {
               int megs = atoi(value);
               if (megs < 1) megs = 1;
               LogTime(MUSCLE_LOG_INFO, "Limiting memory usage to %i megabyte%s.\n", megs, (megs==1)?"":"s");
               maxBytes = megs*1024L*1024L;
            }
            else if (strcmp(keyword, "display") == 0)
            {
               int ll = ParseLogLevelKeyword(value);
               if (ll >= 0) SetConsoleLogLevel(ll);
                       else LogTime(MUSCLE_LOG_INFO, "Error, unknown display log level type [%s]\n", value);
            }
            else if (strcmp(keyword, "log") == 0)
            {
               int ll = ParseLogLevelKeyword(value);
               if (ll >= 0) SetFileLogLevel(ll);
                       else LogTime(MUSCLE_LOG_INFO, "Error, unknown file log level type [%s]\n", value);
            }
            else if (strcmp(keyword, "ban") == 0)
            {
               LogTime(MUSCLE_LOG_INFO, "Banning clients whose IP addresses match [%s]\n", value);   
               tempPrivs.AddString("banned", value);
            }
            else if (strncmp(keyword, "priv", 4) == 0)
            {
               const char * privNames[] = {"kick", "ban", "unban", "all"};
               const char * privType = &keyword[4];
               for (int p=0; p<=PR_NUM_PRIVILEGES; p++)  // if (p == PR_NUM_PRIVILEGES), that means all privileges
               {
                  if (strcmp(privType, privNames[p]) == 0)
                  {
                     LogTime(MUSCLE_LOG_INFO, "Clients whose IP addresses match [%s] get %s privileges\n", value, privNames[p]);
                     char tt[32]; 
                     sprintf(tt, "priv%i", p);
                     tempPrivs.AddString(tt, value);
                  }
               }
            }
            else LogTime(MUSCLE_LOG_WARNING, "Unknown argument [%s] not parsed.\n", keyword);
         }
         else 
         {
            // legacy parsing support
            if ((keyword[0] >= '0')&&(keyword[0] <= '9')) port = atoi(keyword);
            else LogTime(MUSCLE_LOG_WARNING, "Argument [%s] not parsed.  Run ./muscled help for more info.\n", keyword);
         }
      }
   }

   AddOutOfMemoryHandler((OutOfMemoryHandlerFunc)(StorageReflectSession::DrainPools));

   int retVal = 0;
   ReflectServer server(maxBytes);
   PortableMessageFieldNameIterator iter = tempPrivs.GetFieldNameIterator();
   const char * fn;
   while((fn = iter.GetNextFieldName()) != NULL) tempPrivs.CopyName(fn, server.GetCentralState());

   if (server.ServerProcessLoop(StorageSessionCreationFunc, (port > 0) ? port : DEFAULT_MUSCLED_PORT) == B_NO_ERROR) LogTime(MUSCLE_LOG_INFO, "Server process exiting\n");
   else 
   {
      LogTime(MUSCLE_LOG_CRITICALERROR, "Server process aborted\n");
      retVal = 10;
   }
   server.Cleanup();
   return retVal;
}
