/* This file is Copyright 2001 Level Control Systems.  See the included LICENSE.TXT file for details. */

#ifndef MUSCLE_SYS_LOG_H
#define MUSCLE_SYS_LOG_H

#include <stdarg.h>
#include "support/PortableTypeDefs.h"

/** log level constants to use with SetLogLevel(), GetLogLevel() */
enum
{
   MUSCLE_LOG_NONE = 0,         // nothing ever gets logged at this level (default)
   MUSCLE_LOG_CRITICALERROR,    // only things that should never ever happen
   MUSCLE_LOG_ERROR,            // things that shouldn't usually happen
   MUSCLE_LOG_WARNING,          // things that are suspicious
   MUSCLE_LOG_INFO,             // things that the user might like to know
   MUSCLE_LOG_DEBUG,            // things the programmer is debugging with
   MUSCLE_LOG_TRACE,            // exhaustively detailed output
   NUM_MUSCLE_LOGLEVELS
}; 

/** Returns the MUSCLE_LOG_* equivalent of the given keyword string 
 *  @param keyword a string such as "debug", "log", or "info".
 *  @return A MUSCLE_LOG_* value
 */
int ParseLogLevelKeyword(const char * keyword);

/** Returns the current log level for logging to a file.
 *  @return a MUSCLE_LOG_* value.
 */
int GetFileLogLevel();

/** Returns the current log level for logging to stdout.
 *  @return a MUSCLE_LOG_* value.
 */
int GetConsoleLogLevel();

/** Returns the max of GetFileLogLevel() and GetConsoleLogLevel()
 *  @return a MUSCLE_LOG_* value
 */
int GetMaxLogLevel();

/** Sets the log filter level for logging to a file.  
 *  Any calls to Log*() that specify a log level greater than (loglevel)
 *  will be suppressed.  Default level is MUSCLE_LOG_NONE (i.e. no file logging is done)
 *  @param loglevel The MUSCLE_LOG_* value to use in determining which log messages to save to disk.
 */
void SetFileLogLevel(int loglevel);

/** Sets the log filter level for logging to stdout.
 *  Any calls to Log*() that specify a log level greater than (loglevel)
 *  will be suppressed.  Default level is MUSCLE_LOG_INFO.
 *  @param loglevel The MUSCLE_LOG_* value to use in determining which log messages to print to stdout.
 */
void SetConsoleLogLevel(int loglevel);

/** Same semantics as printf, only outputs to the log file/console instead
 *  @param minLogLevel a MUSCLE_LOG_* value indicating the "severity" of this message.
 *  @fmt A printf-style format string (e.g. "hello %s\n").  Note that \n is NOT added for you.
 */
void Log(int minLogLevel, const char * fmt, ...);

/** Formatted.  Automagically prepends a timestamp and status string to your string.
 *  e.g. LogTime(MUSCLE_LOG_INFO, "Hello %s!", "world") would generate "[I 12/18 12:11:49] Hello world!"
 *  @param minLogLevel a MUSCLE_LOG_* value indicating the "severity" of this message.
 *  @fmt A printf-style format string (e.g. "hello %s\n").  Note that \n is NOT added for you.
 */
void LogTime(int minLogLevel, const char * fmt, ...);

/** Returns a human-readable string for the given log level.
 *  @param logLevel A MUSCLE_LOG_* value
 *  @return A pretty human-readable description string such as "Informational" or "Warnings and Errors Only"
 */
const char * GetLogLevelName(int logLevel);

/** Callback object that can be added with AddLogCallback() 
 *  Whenever a log message is generated, all added LogCallback
 *  objects will ahve their Log() methods called.
 */
class LogCallback
{
public:
   /** Destructor, to keep C++ honest */
   virtual ~LogCallback() {/* empty */}

   /** Callback method.  Called whenever a message is logged with Log() or LogTime().
    *  @param when Timestamp of when the message was logged
    *  @param logLevel the MUSCLE_LOG_* value the message was logged with.
    *  @param format the printf-style format string
    *  @param argList the varargs argument list 
    */
   virtual void Log(time_t when, int logLevel, const char * format, va_list argList) = 0;
};

/** Add a custom LogCallback object to the global log callbacks list.
 *  @param cb a dynamically allocated LogCallback object.  The SysLog assumes responsibility for deleting this.
 */
void AddLogCallback(LogCallback * cb);

/** Removes the given callback from our list.  
 *  @param cb The LogCallback to remove from the callback list, or NULL to remove ALL LogCallbacks.
 *  @param deleteToo.  If true, any removed callback(s) will be deleted as well.
 */
void RemoveLogCallback(LogCallback * cb, bool deleteToo);

#endif
